# gen_driver.py
from __future__ import absolute_import
 
from fortlab.kgplugin import Kgen_Plugin
from fortlab.resolver import statements, block_statements, typedecl_statements
from .gencore_utils import DRIVER_USE_PART, DRIVER_READ_IN_ARGS, DRIVER_CALLSITE_PART, DRIVER_DECL_PART, \
    DRIVER_EXEC_PART, DRIVER_CONTAINS_PART, DRIVER_SUBP_PART, DRIVER_ALLOC_PART, DRIVER_DEALLOC_PART, \
    shared_objects, DRIVER_READ_IN_EXTERNS, DRIVER_WARMUP_ALLOC_PART, DRIVER_WARMUP_READ_IN_ARGS, \
    DRIVER_WARMUP_READ_IN_EXTERNS, DRIVER_WARMUP_CALLSITE_PART, DRIVER_WARMUP_DEALLOC_PART, \
    DRIVER_EVAL_ALLOC_PART, DRIVER_EVAL_READ_IN_ARGS, \
    DRIVER_EVAL_READ_IN_EXTERNS, DRIVER_EVAL_CALLSITE_PART, DRIVER_EVAL_DEALLOC_PART


class Gen_K_Driver(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial event
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
            block_statements.Program, self.is_driver_name, self.create_kernel_driver_parts) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.FINISH_PROCESS, \
            block_statements.Program, self.is_driver_name, self.finalize_kernel_driver_parts) 
    # match functions
    def is_driver_name(self, node):
        if node.name==getinfo('kernel_driver_name'): return True
        else: return False

    def finalize_kernel_driver_parts(self, node):

        # add callsite stmts
        attrs = {'designator': getinfo('parentblock_stmt').name, 'items': getinfo('kernel_driver_callsite_args')}
        namedpart_append_genknode(node.kgen_kernel_id, DRIVER_CALLSITE_PART, statements.Call, attrs=attrs)

        for partname, wpartname in [ (DRIVER_ALLOC_PART, DRIVER_WARMUP_ALLOC_PART), (DRIVER_READ_IN_ARGS, DRIVER_WARMUP_READ_IN_ARGS), \
            (DRIVER_READ_IN_EXTERNS, DRIVER_WARMUP_READ_IN_EXTERNS), (DRIVER_CALLSITE_PART, DRIVER_WARMUP_CALLSITE_PART), \
            (DRIVER_DEALLOC_PART, DRIVER_WARMUP_DEALLOC_PART) ]:
            pnode, rawname, named_part =  get_namedpart(node.kgen_kernel_id, partname)
            for item in named_part:
                namedpart_append_node(node.kgen_kernel_id, wpartname, item)

        for partname, epartname in [ (DRIVER_ALLOC_PART, DRIVER_EVAL_ALLOC_PART), (DRIVER_READ_IN_ARGS, DRIVER_EVAL_READ_IN_ARGS), \
            (DRIVER_READ_IN_EXTERNS, DRIVER_EVAL_READ_IN_EXTERNS), (DRIVER_CALLSITE_PART, DRIVER_EVAL_CALLSITE_PART), \
            (DRIVER_DEALLOC_PART, DRIVER_EVAL_DEALLOC_PART) ]:
            pnode, rawname, named_part =  get_namedpart(node.kgen_kernel_id, partname)
            for item in named_part:
                namedpart_append_node(node.kgen_kernel_id, epartname, item)

    #  process after node creation
    def create_kernel_driver_parts(self, node):

        shared_objects['driver_object'] = node

        namedpart_link_part(node, DRIVER_USE_PART, USE_PART)
        namedpart_link_part(node, DRIVER_DECL_PART, DECL_PART)
        namedpart_link_part(node, DRIVER_EXEC_PART, EXEC_PART)
        namedpart_link_part(node, DRIVER_CONTAINS_PART, CONTAINS_PART)
        namedpart_link_part(node, DRIVER_SUBP_PART, SUBP_PART)

        attrs = {'name':'kgen_utils_mod', 'isonly': True, 'items':['kgen_get_newunit', 'kgen_error_stop', 'kgen_dp', 'kgen_array_sumcheck', 'kgen_rankthreadinvoke']}
        part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)

        #attrs = {'name':'tprof_mod', 'isonly': True, 'items':['tstart', 'tstop', 'tnull', 'tprnt']}
        #part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)

        #attrs = {'name':'IEEE_ARITHMETIC', 'nature': 'INTRINSIC', 'isonly': True, 'items':['ieee_is_normal']}
        #part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)

        attrs = {'name':getinfo('topblock_stmt').name, 'isonly': True, 'items':[getinfo('parentblock_stmt').name]}
        part_append_genknode(node, USE_PART, statements.Use, attrs=attrs)
        part_append_comment(node, USE_PART, '')

        part_append_genknode(node, IMPLICIT_PART, typedecl_statements.Implicit)
        part_append_comment(node, IMPLICIT_PART, '')

        if getinfo('add_mpi_frame'):
            part_append_comment(node, DECL_PART, '#ifdef _MPI', style='rawtext')
            part_append_comment(node, DECL_PART, 'include "mpif.h"', style='rawtext')
            part_append_comment(node, DECL_PART, '#endif', style='rawtext')
            part_append_comment(node, DECL_PART, '')

        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_isverified']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_ierr_list', 'kgen_unit_list']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_ierr', 'kgen_unit', 'kgen_case_count', 'kgen_count_verified']}
        part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'entity_decls': ['kgen_filepath'], 'selector':('1024', None)}
        part_append_genknode(node, DECL_PART, typedecl_statements.Character, attrs=attrs)

        if getinfo('is_papi_enabled'):
            part_append_comment(node, DECL_PART, '#ifdef KGEN_PAPI', style='rawtext')

            attrs = {'type_spec': 'INTEGER', 'selector': (None, '8'), 'entity_decls': \
                ['kgen_measure', 'kgen_total_count', 'kgen_min_count', 'kgen_max_count'], }
            part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            part_append_comment(node, DECL_PART, '#else', style='rawtext')

        attrs = {'type_spec': 'REAL', 'selector': (None, 'kgen_dp'), 'entity_decls': \
            ['kgen_measure', 'kgen_total_time', 'kgen_min_time', 'kgen_max_time'], }
        part_append_genknode(node, DECL_PART, typedecl_statements.Real, attrs=attrs)

        if getinfo('is_papi_enabled'):
            part_append_comment(node, DECL_PART, '#endif', style='rawtext')

        attrs = {'type_spec': 'REAL', 'entity_decls': ['kgen_array_sum'], 'selector': (None, '8')}
        part_append_genknode(node, DECL_PART, typedecl_statements.Real, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_mpirank', 'kgen_openmptid', 'kgen_kernelinvoke']}
        part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['myrank', 'mpisize']}
        part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_evalstage', 'kgen_warmupstage', 'kgen_mainstage']}
        part_append_gensnode(node, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'items': [ ( 'kgen_state', ('kgen_mpirank', 'kgen_openmptid', 'kgen_kernelinvoke', 'kgen_evalstage', 'kgen_warmupstage', 'kgen_mainstage') ) ]}
        part_append_gensnode(node, DECL_PART, statements.Common, attrs=attrs)

        part_append_comment(node, DECL_PART, '')

        if getinfo('add_mpi_frame'):
            part_append_comment(node, EXEC_PART, '#ifdef _MPI', style='rawtext')
            part_append_comment(node, EXEC_PART, 'CALL MPI_INIT(kgen_ierr)', style='rawtext')
            part_append_comment(node, EXEC_PART, 'IF (kgen_ierr .NE. MPI_SUCCESS) THEN', style='rawtext')
            part_append_comment(node, EXEC_PART, '  PRINT *, "MPI Initialization is failed."', style='rawtext')
            part_append_comment(node, EXEC_PART, '  CALL MPI_ABORT(MPI_COMM_WORLD, -1, kgen_ierr)', style='rawtext')
            part_append_comment(node, EXEC_PART, 'END IF', style='rawtext')
            part_append_comment(node, EXEC_PART, 'call mpi_comm_rank(mpi_comm_world, myrank, kgen_ierr)', style='rawtext')
            part_append_comment(node, EXEC_PART, 'call mpi_comm_size(mpi_comm_world, mpisize, kgen_ierr)', style='rawtext')
            part_append_comment(node, EXEC_PART, '#else', style='rawtext')

        part_append_comment(node, EXEC_PART, 'myrank = 0', style='rawtext')
        part_append_comment(node, EXEC_PART, 'mpisize = 1', style='rawtext')

        if getinfo('add_mpi_frame'):
            part_append_comment(node, EXEC_PART, '#endif', style='rawtext')
            part_append_comment(node, EXEC_PART, '')


        if getinfo('is_papi_enabled'):
            part_append_comment(node, EXEC_PART, '#ifdef KGEN_PAPI', style='rawtext')

            attrs = {'variable': 'kgen_total_count', 'sign': '=', 'expr': '0_8'}
            part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_min_count', 'sign': '=', 'expr': 'HUGE(0_8)'}
            part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_max_count', 'sign': '=', 'expr': '0_8'}
            part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)

            part_append_comment(node, EXEC_PART, '#else', style='rawtext')

        attrs = {'variable': 'kgen_total_time', 'sign': '=', 'expr': '0.0_kgen_dp'}
        part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_min_time', 'sign': '=', 'expr': 'HUGE(0.0_kgen_dp)'}
        part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_max_time', 'sign': '=', 'expr': '0.0_kgen_dp'}
        part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)

        if getinfo('is_papi_enabled'):
            part_append_comment(node, EXEC_PART, '#endif', style='rawtext')

        attrs = {'variable': 'kgen_case_count', 'sign': '=', 'expr': '0'}
        part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_count_verified', 'sign': '=', 'expr': '0'}
        part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_comment(node, EXEC_PART, '')

        attrs = {'variable': 'kgen_unit_list', 'sign': '=', 'expr': 'kgen_get_newunit()'}
        part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)
      
        attrs = {'specs': ['UNIT=kgen_unit_list', 'FILE="kgen_statefile.lst"', 'STATUS="OLD"', 'IOSTAT=kgen_ierr_list']}
        part_append_genknode(node, EXEC_PART, statements.Open, attrs=attrs)

        attrs = {'expr': 'kgen_ierr_list .NE. 0'}
        iflist = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'designator': 'SYSTEM', 'items': ['"ls -1 %s.*.*.* > kgen_statefile.lst"'%getinfo('kernel_name')]}
        part_append_genknode(iflist, EXEC_PART, statements.Call, attrs=attrs)

        attrs = {'designator': 'SLEEP', 'items': ['1']}
        part_append_genknode(iflist, EXEC_PART, statements.Call, attrs=attrs)

        attrs = {'variable': 'kgen_unit_list', 'sign': '=', 'expr': 'kgen_get_newunit()'}
        part_append_genknode(iflist, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'specs': ['UNIT=kgen_unit_list', 'FILE="kgen_statefile.lst"', 'STATUS="OLD"', 'IOSTAT=kgen_ierr_list']}
        part_append_genknode(iflist, EXEC_PART, statements.Open, attrs=attrs)

        attrs = {'expr': 'kgen_ierr_list .NE. 0'}
        iflist2 = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'myrank == 0'}
        ifrank = part_append_genknode(iflist2, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"ERROR: ""kgen_statefile.lst"" is not found in current directory."']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        part_append_genknode(iflist2, EXEC_PART, statements.Stop)

        # eval
        attrs = {'loopcontrol': 'WHILE ( kgen_ierr_list .EQ. 0 )'}
        doobj1 = part_append_genknode(node, EXEC_PART, block_statements.Do, attrs=attrs)

        attrs = {'items': ['kgen_filepath'], 'specs': ['UNIT = kgen_unit_list', 'FMT="(A)"', 'IOSTAT=kgen_ierr_list']}
        part_append_genknode(doobj1, EXEC_PART, statements.Read, attrs=attrs)

        attrs = {'expr': 'kgen_ierr_list .EQ. 0'}
        ifread = part_append_genknode(doobj1, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'kgen_unit', 'sign': '=', 'expr': 'kgen_get_newunit()'}
        part_append_genknode(ifread, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'designator': 'kgen_rankthreadinvoke', 'items': ( 'TRIM(ADJUSTL(kgen_filepath))', 'kgen_mpirank', 'kgen_openmptid', 'kgen_kernelinvoke' ) }
        part_append_genknode(ifread, EXEC_PART, statements.Call, attrs=attrs)

        attrs = {'specs': ['UNIT=kgen_unit', 'FILE=TRIM(ADJUSTL(kgen_filepath))', 'STATUS="OLD"', 'ACCESS="STREAM"', \
            'FORM="UNFORMATTED"', 'ACTION="READ"', 'CONVERT="BIG_ENDIAN"', 'IOSTAT=kgen_ierr']}
        part_append_genknode(ifread, EXEC_PART, statements.Open, attrs=attrs)

        attrs = {'expr': 'kgen_ierr == 0'}
        ifopen = part_append_genknode(ifread, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'myrank == 0'}
        ifrank = part_append_genknode(ifopen, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"***************** Verification against \'" // trim(adjustl(kgen_filepath)) // "\' *****************"']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'variable': 'kgen_evalstage', 'sign': '=', 'expr': '.TRUE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_warmupstage', 'sign': '=', 'expr': '.FALSE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_mainstage', 'sign': '=', 'expr': '.FALSE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_comment(ifopen, EXEC_PART, '')

        # register gencore parts
        namedpart_create_subpart(ifopen, DRIVER_EVAL_ALLOC_PART, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_EVAL_READ_IN_ARGS, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_EVAL_READ_IN_EXTERNS, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_EVAL_CALLSITE_PART, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_EVAL_DEALLOC_PART, EXEC_PART)

        attrs = {'specs': ['UNIT=kgen_unit']}
        part_append_genknode(ifopen, EXEC_PART, statements.Rewind, attrs=attrs)

        # warm up
        attrs = {'variable': 'kgen_evalstage', 'sign': '=', 'expr': '.FALSE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_warmupstage', 'sign': '=', 'expr': '.TRUE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_mainstage', 'sign': '=', 'expr': '.FALSE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_comment(ifopen, EXEC_PART, '')

        # register gencore parts
        namedpart_create_subpart(ifopen, DRIVER_WARMUP_ALLOC_PART, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_WARMUP_READ_IN_ARGS, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_WARMUP_READ_IN_EXTERNS, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_WARMUP_CALLSITE_PART, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_WARMUP_DEALLOC_PART, EXEC_PART)

        attrs = {'specs': ['UNIT=kgen_unit']}
        part_append_genknode(ifopen, EXEC_PART, statements.Rewind, attrs=attrs)

        # main
        attrs = {'variable': 'kgen_evalstage', 'sign': '=', 'expr': '.FALSE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_warmupstage', 'sign': '=', 'expr': '.FALSE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_mainstage', 'sign': '=', 'expr': '.TRUE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        #attrs = {'items': ['"***************** Verification against \'" // trim(adjustl(kgen_filepath)) // "\' *****************"']}
        #part_append_genknode(ifopen, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'variable': 'kgen_case_count', 'sign': '=', 'expr': 'kgen_case_count + 1'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_isverified', 'sign': '=', 'expr': '.FALSE.'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_comment(ifopen, EXEC_PART, '')

        # register gencore parts
        namedpart_create_subpart(ifopen, DRIVER_ALLOC_PART, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_READ_IN_ARGS, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_READ_IN_EXTERNS, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_CALLSITE_PART, EXEC_PART)
        namedpart_create_subpart(ifopen, DRIVER_DEALLOC_PART, EXEC_PART)


        namedpart_append_comment(node.kgen_kernel_id, DRIVER_READ_IN_ARGS, '')
        namedpart_append_comment(node.kgen_kernel_id, DRIVER_READ_IN_ARGS, 'driver read in arguments')

        namedpart_append_comment(node.kgen_kernel_id, DRIVER_READ_IN_EXTERNS, '')
        namedpart_append_comment(node.kgen_kernel_id, DRIVER_READ_IN_EXTERNS, 'extern input variables')

        namedpart_append_comment(node.kgen_kernel_id, DRIVER_CALLSITE_PART, '')
        namedpart_append_comment(node.kgen_kernel_id, DRIVER_CALLSITE_PART, 'callsite part')


        if getinfo('is_papi_enabled'):
            part_append_comment(ifopen, EXEC_PART, '#ifdef KGEN_PAPI', style='rawtext')

            attrs = {'variable': 'kgen_total_count', 'sign': '=', 'expr': 'kgen_total_count + kgen_measure'}
            part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_min_count', 'sign': '=', 'expr': 'MIN( kgen_min_count, kgen_measure )'}
            part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'variable': 'kgen_max_count', 'sign': '=', 'expr': 'MAX( kgen_max_count, kgen_measure )'}
            part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

            part_append_comment(ifopen, EXEC_PART, '#else', style='rawtext')

        attrs = {'variable': 'kgen_total_time', 'sign': '=', 'expr': 'kgen_total_time + kgen_measure'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_min_time', 'sign': '=', 'expr': 'MIN( kgen_min_time, kgen_measure )'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'variable': 'kgen_max_time', 'sign': '=', 'expr': 'MAX( kgen_max_time, kgen_measure )'}
        part_append_genknode(ifopen, EXEC_PART, statements.Assignment, attrs=attrs)

        if getinfo('is_papi_enabled'):
            part_append_comment(ifopen, EXEC_PART, '#endif', style='rawtext')

        attrs = {'expr': 'kgen_isverified'}
        ifverified = part_append_genknode(ifopen, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'kgen_count_verified', 'sign': '=', 'expr': 'kgen_count_verified + 1'}
        part_append_genknode(ifverified, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'specs': ['UNIT=kgen_unit']}
        part_append_genknode(ifread, EXEC_PART, statements.Close, attrs=attrs)

        part_append_comment(node, EXEC_PART, '')

        attrs = {'specs': ['UNIT=kgen_unit_list']}
        part_append_genknode(node, EXEC_PART, statements.Close, attrs=attrs)

        part_append_comment(node, EXEC_PART, '')

        attrs = {'expr': 'myrank == 0'}
        ifrank = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"****************************************************"'], 'specs': [ '*', '"(A)"' ]}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"kernel execution summary: %s"'%getinfo('kernel_name')], 'specs': [ '*', '"(4X,A)"' ]}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"****************************************************"'], 'specs': [ '*', '"(A)"' ]}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'expr': 'kgen_case_count == 0'}
        ifcount = part_append_genknode(ifrank, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['"No data file is verified."'] }
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        part_append_genknode(ifcount, EXEC_PART, statements.Else)

        attrs = {'items': ['"Total number of verification cases  "', '":"', 'kgen_case_count'], 'specs': [ '*', '"(4X, A36, A1, I6)"' ]}
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Number of verification-passed cases "', '":"', 'kgen_count_verified'], 'specs': [ '*', '"(4X, A36, A1, I6)"' ]}
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        #attrs = {'items': ['"Verification tolerance "', '":"', '%s'%getinfo('verify_tol')], 'specs': [ '*', '"(4X, A22, A1, E24.16)"' ]}
        #part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'expr': 'kgen_case_count == kgen_count_verified'}
        ifpass = part_append_genknode(ifcount, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['"kernel: %s: PASSED verification"'%getinfo("kernel_name")], 'specs': [ '*', '"(4X,A)"' ]}
        part_append_genknode(ifpass, EXEC_PART, statements.Write, attrs=attrs)

        part_append_genknode(ifpass, EXEC_PART, statements.Else)

        attrs = {'items': ['"kernel: %s: FAILED verification"'%getinfo("kernel_name")], 'specs': [ '*', '"(4X,A)"' ]}
        part_append_genknode(ifpass, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        if getinfo('is_papi_enabled'):
            part_append_comment(ifcount, EXEC_PART, '#ifdef KGEN_PAPI', style='rawtext')

            attrs = {'items': ['"Average KGENPAPIEVENT: ", NINT(kgen_total_count / DBLE(kgen_case_count))'], 'specs': [ '*', '"(4X, A, I16)"' ]}
            part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"Minimum KGENPAPIEVENT: ", kgen_min_count'], 'specs': [ '*', '"(4X, A, I16)"' ]}
            part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"Maximum KGENPAPIEVENT: ", kgen_max_count'], 'specs': [ '*', '"(4X, A, I16)"' ]}
            part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

            part_append_comment(ifcount, EXEC_PART, '#else', style='rawtext')

        attrs = {'items': ['"number of processes: ", mpisize'], 'specs': [ '*', '"(4X,A19,I3)"' ]}
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['""']}
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Average call time (usec): ", kgen_total_time / DBLE(kgen_case_count)'], 'specs': [ '*', '"(4X, A, E10.3)"' ]}
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Minimum call time (usec): ", kgen_min_time'], 'specs': [ '*', '"(4X, A, E10.3)"' ]}
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'items': ['"Maximum call time (usec): ", kgen_max_time'], 'specs': [ '*', '"(4X, A, E10.3)"' ]}
        part_append_genknode(ifcount, EXEC_PART, statements.Write, attrs=attrs)

        if getinfo('is_papi_enabled'):
            part_append_comment(ifcount, EXEC_PART, '#endif', style='rawtext')

        attrs = {'items': ['"****************************************************"'], 'specs': [ '*', '"(A)"' ]}
        part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

        if getinfo('add_mpi_frame'):
            part_append_comment(node, EXEC_PART, '')
            part_append_comment(node, EXEC_PART, '#ifdef _MPI', style='rawtext')
            part_append_comment(node, EXEC_PART, 'CALL mpi_finalize(kgen_ierr)', style='rawtext')
            part_append_comment(node, EXEC_PART, '#endif', style='rawtext')
            part_append_comment(node, EXEC_PART, '')

        # Block Data
        attrs = {'name': 'KGEN'}
        cblock = part_append_gensnode(node.kgen_parent, UNIT_PART, block_statements.BlockData, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['kgen_mpirank = 0', 'kgen_openmptid = 0', 'kgen_kernelinvoke = 0']}
        part_append_gensnode(cblock, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['kgen_evalstage = .TRUE.', 'kgen_warmupstage = .FALSE.', 'kgen_mainstage = .FALSE.']}
        part_append_gensnode(cblock, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'items': [ ( 'kgen_state', ('kgen_mpirank', 'kgen_openmptid', 'kgen_kernelinvoke', 'kgen_evalstage', 'kgen_warmupstage', 'kgen_mainstage') ) ]}
        part_append_gensnode(cblock, DECL_PART, statements.Common, attrs=attrs)




'''KGen parser
'''
import os, io, locale
#import os.path
from fortlab.kgutils import UserException, pack_exnamepath, match_namepath, traverse, run_shcmd, logger
from fortlab.resolver.statements import Comment
from fortlab.resolver.block_statements import Module, Program
from fortlab.resolver import api
from collections import OrderedDict

#import logging
#logger = logging.getLogger('kgen')

#############################################################################
## RESOLUTION TYPE
#############################################################################

class KGGenType(object):
    STATE_IN = 0x2
    STATE_OUT = 0x3

    @classmethod
    def is_state_in(cls, value):
        return cls.STATE_IN == value

    @classmethod
    def is_state_out(cls, value):
        return cls.STATE_OUT == value

    @classmethod
    def is_state(cls, value):
        return cls.is_state_in(value) or cls.is_state_out(value)

    @classmethod
    def has_state_in(cls, geninfo):
        return cls.STATE_IN in geninfo

    @classmethod
    def has_state_out(cls, geninfo):
        return cls.STATE_OUT in geninfo

    @classmethod
    def has_state(cls, geninfo):
        return cls.has_state_in(geninfo) or cls.has_state_out(geninfo)

    @classmethod
    def get_state_in(cls, geninfo):
        return geninfo.get(cls.STATE_IN, [])

    @classmethod
    def get_state_out(cls, geninfo):
        return geninfo.get(cls.STATE_OUT, [])

    @classmethod
    def get_state(cls, geninfo):
        state = cls.get_state_in(geninfo)
        #state = cls.get_state_in_inout(geninfo)
        for uname, req in cls.get_state_out(geninfo):
            if all(not uname==u for u, r in state):
                state.append((uname, req))
        return state

    @classmethod
    def get_request_in(cls, uname, geninfo):
        if cls.has_state_in(geninfo):
            for sin_uname, req in cls.get_state_in(geninfo):
                if uname==sin_uname: return req

    @classmethod
    def get_request_out(cls, uname, geninfo):
        if cls.has_state_out(geninfo):
            for sout_uname, req in cls.get_state_out(geninfo):
                if uname==sout_uname: return req

    @classmethod
    def get_request(cls, uname, geninfo):
        if cls.has_state(geninfo):
            for s_uname, req in cls.get_state(geninfo):
                if uname==s_uname: return req

    @classmethod
    def has_uname_in(cls, uname, geninfo):
        if cls.get_request_in(uname, geninfo): return True
        else: return False

    @classmethod
    def has_uname_out(cls, uname, geninfo):
        if cls.get_request_out(uname, geninfo): return True
        else: return False

    @classmethod
    def has_uname(cls, uname, geninfo):
        if cls.get_request(uname, geninfo): return True
        else: return False

class ResState(object):
    ( NOT_STARTED, RESOLVED ) = range(2)

    def __init__(self, gentype, uname, org, resolvers):
        self.state = self.NOT_STARTED
        self.gentype = gentype
        self.uname = uname
        self.originator = org
        self.resolvers = resolvers
        self.res_stmts = []
        self.unamelist = [uname]

    def push_uname(self, uname):
        self.unamelist.append(uname)
        self.uname = uname

    def pop_uname(self, reset_uname=False):
        newname = self.unamelist.pop()
        self.uname = self.unamelist[-1]
        if len(self.res_stmts)>0 and reset_uname:
            newlist = []
            values = list(self.res_stmts[-1].geninfo.values())
            #for (resuname, req) in self.res_stmts[-1].geninfo.values()[0]:
            for (resuname, req) in values[0]:
                if resuname==newname:
                    newlist.append((self.uname, req))
                else:
                    newlist.append((resuname, req))
                    pass
            #self.res_stmts[-1].geninfo.values()[0] = newlist
            values[0] = newlist

class SrcFile(object):
    def handle_include(self, lines):
        import re
        import os

        insert_lines = []
        for i, line in enumerate(lines):
            match = re.match(r'^\s*include\s*("[^"]+"|\'[^\']+\')', line, re.I)
            #if not match:
            #    match = re.match(r'\s*#include\s*("[^"]+"|\<[^\']+\>)\s*\Z', line, re.I)
            if match:
                if self.realpath in self.config["include"]['file']:
                    include_dirs = (self.config["include"]['file'][self.realpath]['path'] +
                                    self.config["include"]['path'])
                else:
                    include_dirs = self.config["include"]['path']

                if os.path.isfile(self.config["mpi"]['header']):
                    include_dirs.insert(0, os.path.dirname(self.config["mpi"]['header']))

                filename = match.group(1)[1:-1].strip()
                path = filename
                for incl_dir in include_dirs+[os.path.dirname(self.realpath)]:
                    path = os.path.join(incl_dir, filename)
                    if os.path.exists(path):
                        break
                if os.path.isfile(path):
                    with open(path, 'r') as f:
                        included_lines = f.read()
                        insert_lines.extend(self.handle_include(
                                            included_lines.split('\n')))
                else:
                    raise UserException('Can not find %s in include paths of %s.' %
                                        (filename, self.realpath))
            else:
                insert_lines.append(line)

        return insert_lines

    def __init__(self, srcpath, config, preprocess=True):

        # set default values
        self.tree = None
        self.config = config
        self.srcpath = srcpath
        self.realpath = os.path.realpath(self.srcpath)
        
        # set source file format
        isfree = None
        isstrict = None
        if self.realpath in config["source"]['file'].keys():
            if 'isfree' in config["source"]['file'][self.realpath]:
                isfree = config["source"]['file'][self.realpath]['isfree']
            if 'isstrict' in config["source"]['file'][self.realpath]:
                isstrict = config["source"]['file'][self.realpath]['isstrict']
        else:
            isstrict = config["source"]['isstrict']
            isfree = config["source"]['isfree']

        # prepare include paths and macro definitions
        path_src = []
        macros_src = []
        if self.realpath in config["include"]['file']:
            rpath = config["include"]['file'][self.realpath]
            srcbackup = rpath['srcbackup']
            path_src = rpath['path'] + [os.path.dirname(self.realpath)]
            path_src = [path for path in path_src if len(path)>0]
            for k, v in rpath['macro'].items():
                if v is not None:
                    macros_src.append('-D%s=%s'%(k,v))
                else:
                    macros_src.append('-D%s'%k)
        else:
            srcbackup = []

        if os.path.isfile(config["mpi"]['header']):
            includes = ['-I %s'%incpath for incpath in
                        ([os.path.dirname(config["mpi"]['header'])] +
                        config["include"]['path']+path_src)]
        else:
            includes = ['-I %s'%incpath for incpath in
                        config["include"]['path']+path_src ]

        macros_common = []
        for k, v in config["include"]['macro'].items():
            if v:
                macros_common.append('-D%s=%s'%(k,v))
            else:
                macros_common.append('-D%s'%k)
        macros = ' '.join(macros_common + macros_src)

        # execute preprocessing
        logger.info('Reading %s'%self.srcpath)

        new_lines = []

        enc = locale.getpreferredencoding(False)

        opensrcpath = self.realpath

        if not os.path.isfile(opensrcpath) and srcbackup:
            opensrcpath = srcbackup[0]

        with io.open(opensrcpath, 'r', encoding=enc) as f:
            if preprocess:
                pp = config["bin"]['pp']
                if pp.endswith('fpp'):
                    if isfree is None or isfree: srcfmt = ' -free'
                    else: srcfmt = ' -fixed'
                    flags = config["bin"]['fpp_flags'] + srcfmt
                elif pp.endswith('cpp'):
                    flags = config["bin"]['cpp_flags']
                else: raise UserException(
                                'Preprocessor is not either fpp or cpp')

                cmd = '%s %s %s %s' % (pp, flags, ' '.join(includes), macros)
                output, err, retcode = run_shcmd(cmd, input=f.read())
                output = output.decode("utf-8")
                prep = map(lambda l: '!KGEN'+l if l.startswith('#') else l,
                           output.split('\n'))
                new_lines = self.handle_include(prep)
            else:
                new_lines = f.read().split('\n')

        # add include paths
        include_dirs = config["include"]['path'][:]
        ifile = config["include"]['file']
        if (self.realpath in ifile and
            "path" in ifile[self.realpath]):
            include_dirs.extend(ifile[self.realpath]['path'])
            include_dirs.append(os.path.dirname(self.realpath))

        # fparse

        #if self.realpath.endswith("ESMF_ShrTimeMod.F90"):
        #    import pdb; pdb.set_trace()

        self.tree = api.parse('\n'.join(new_lines), ignore_comments=False,
                              analyze=True, isfree=isfree, isstrict=isstrict,
                              include_dirs=include_dirs, source_only=None )
        self.tree.prep = new_lines

        # parse f2003
        lineno = 0
        linediff = 0
        for stmt, depth in api.walk(self.tree, -1):
            stmt.parse_f2003()

        # rename reader.id
        self.tree.reader.id = self.realpath

        # collect module information
        for mod_name, mod_stmt in self.tree.a.module.items(): 
            if mod_name not in config["modules"]:
                config["modules"][mod_name] = OrderedDict()
                config["modules"][mod_name]['stmt'] = mod_stmt
                config["modules"][mod_name]['file'] = self
                config["modules"][mod_name]['path'] = self.realpath
        
        # collect program unit information
        for item in self.tree.content:
            if item.__class__ not in [ Module, Comment, Program ]:
                if item.reader.id not in config["program_units"].keys():
                    config["program_units"][item.reader.id] = []
                config["program_units"][item.reader.id].append(item)

        # create a tuple for file dependency
        config["srcfiles"][self.realpath] = ( self, [], [] )

        self.process_directive(config)

    def stmt_by_name(self, name, cls=None, lineafter=-1):
        from statements import Comment

        for stmt, depth in walk(self.tree, -1):
            if isinstance(cls, list):
                if not stmt.__class__ in cls: continue 

            if lineafter>0:
                if stmt.item.span[1]<=lineafter: continue
                if isinstance(stmt, Comment): continue
 
            expr = stmt.expr_by_name(name, stmt.f2003)
            if expr: return stmt, expr

        return None, None

    def process_directive(self, config):
        from fortlab.resolver.kgsearch import f2003_search_unknowns
        from fortlab.resolver.statements import Comment
        from fortlab.resolver.block_statements import executable_construct
        import re

        def get_next_non_comment(stmt):
            if not stmt: return
            if not hasattr(stmt, 'parent'): return

            started = False
            for s in stmt.parent.content:
                if s==stmt:
                    if not isinstance(s, Comment): return s
                    started = True
                elif started:
                    if not isinstance(s, Comment): return s

        def get_names(node, bag, depth):
            from fortlab.resolver.Fortran2003 import Name
            if isinstance(node, Name) and not node.string in bag:
                bag.append(node.string)
           
        # collect directives
        directs = []
        for stmt, depth in api.walk(self.tree):
            if isinstance(stmt, Comment):
                line = stmt.item.comment.strip()
                match = re.match(r'^[c!*]\$kgen\s+(.+)$', line, re.IGNORECASE)
                if match:
                    dsplit = match.group(1).split(' ', 1)
                    dname = dsplit[0].strip()
                    if len(dsplit)>1: clause = dsplit[1].strip()
                    else: clause = None

                    if dname.startswith('begin_'):
                        sname = dname[6:]
                        directs.append(sname)
                        config["kernel"]['name'] = clause

                        if sname in ('mixed',):
                            match = re.finditer(r'\(', clause)
                            if match:
                                for m in match: 
                                    start = m.start()
                                    varname = clause[:start]
                                    pos = clause.find(")", start)
                                    if pos < 0:
                                        raise Exception("Can not not closing parenthesis")
                                    config["mixed"]['varnames'][varname.strip()] = clause[start+1:pos]

                            else:
                                raise Exception("No variable in mixedprec directive")

                    elif dname.startswith('end_'):
                        ename = dname[4:]
                        if directs[-1]==ename:
                            directs.pop()
                            if ename in ('callsite', 'mixed'):
                                while isinstance(config["callsite"]['stmts'][-1], Comment):
                                    config["callsite"]['stmts'].pop()
                            else:
                                raise UserException('WARNING: Not supported KGEN directive: %s'%ename)
                        else:
                            raise UserException('Directive name mismatch: %s, %s'%(dname_stack[-1], ename))
                    elif dname=='callsite':
                        next_fort_stmt = get_next_non_comment(stmt)
                        if next_fort_stmt:
                            config["kernel"]['name'] = clause
                            config["callsite"]['stmts'].append(next_fort_stmt)
                        else:
                            raise UserException('WARNING: callsite is not found')

                    elif dname=='write':
                        if clause:
                            stmt.write_state = tuple( c.strip() for c in clause.split(',') )
                            if not hasattr(stmt, 'unknowns'):
                                f2003_search_unknowns(stmt, stmt.f2003)
                            if hasattr(stmt, 'unknowns'):
                                for unk, req in stmt.unknowns.items():
                                    if req.state != ResState.RESOLVED:
                                        stmt.resolve(req) 

                    elif dname=='exclude':
                        next_fort_stmt = get_next_non_comment(stmt)
                        if next_fort_stmt:
                            next_fort_stmt.f2003.skip_search = True
                            next_fort_stmt.f2003.after_exclude = True
                        else:
                            raise UserException('WARNING: exclude target is not found')

                    elif dname=='coverage':
                        next_fort_stmt = get_next_non_comment(stmt)
                        if next_fort_stmt:
                            next_fort_stmt.f2003.after_coverage = True
                            next_fort_stmt.f2003.coverage_name = clause
                        else:
                            raise UserException('WARNING: coverage target is not found')

                elif 'callsite' in directs or 'mixed' in directs: # if not match and within callsite
                    if config["callsite"]['stmts'] or not isinstance(stmt, Comment):
                        config["callsite"]['stmts'].append(stmt)
            elif 'callsite' in directs or 'mixed' in directs: # if not Comment
                config["callsite"]['stmts'].append(stmt)
            else: # not in callsite
                if config["callsite"]['namepath'] and stmt.__class__ in executable_construct:
                    names = []
                    traverse(stmt.f2003, get_names, names)
                    for name in names:
                        if match_namepath(config["callsite"]['namepath'], pack_exnamepath(stmt, name), internal=False):
                            config["kernel"]['name'] = name
                            for _s, _d in api.walk(stmt):
                                config["callsite"]['stmts'].append(_s)
                            return
                elif len(directs)>0 and directs[-1]=='callsite':
                    config["callsite"]['stmts'].append(stmt)



#!/usr/bin/env python
"""
Defines LineSplitter and helper functions.

-----
Permission to use, modify, and distribute this software is given under the
terms of the NumPy License. See http://scipy.org.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Author: Pearu Peterson <pearu@cens.ioc.ee>
Created: May 2006
-----
"""

__all__ = ['String','string_replace_map','splitquote','splitparen']

import re

class String(str): pass
class ParenString(str): pass

def split2(line, lower=False):
    """
    Split line into non-string part and into a start of a string part.
    Returns 2-tuple. The second item either is empty string or start
    of a string part.
    """
    return LineSplitter(line,lower=lower).split2()

_f2py_str_findall = re.compile(r"_F2PY_STRING_CONSTANT_\d+_").findall
_is_name = re.compile(r'\w*\Z',re.I).match
_is_simple_str = re.compile(r'\w*\Z',re.I).match
_f2py_findall = re.compile(r'(_F2PY_STRING_CONSTANT_\d+_|F2PY_EXPR_TUPLE_\d+)').findall

class string_replace_dict(dict):
    """
    Dictionary object that is callable for applying map returned
    by string_replace_map() function.
    """
    def __call__(self, line):
        for k in _f2py_findall(line):
            line = line.replace(k, self[k])
        return line

def string_replace_map(line, lower=False,
                       _cache={'index':0,'pindex':0}):
    """
    1) Replaces string constants with symbol '_F2PY_STRING_CONSTANT_<index>_'`
    2) Replaces (expression) with symbol `(F2PY_EXPR_TUPLE_<index>)`
    Returns a new line and the replacement map.
    """

    items = []
    string_map = string_replace_dict()
    rev_string_map = {}
    for item in splitquote(line, lower=lower)[0]:
        if isinstance(item, String) and not _is_simple_str(item[1:-1]):
            key = rev_string_map.get(item)
            if key is None:
                _cache['index'] += 1
                index = _cache['index']
                key = "_F2PY_STRING_CONSTANT_%s_" % (index)
                it = item[1:-1]
                string_map[key] = it
                rev_string_map[it] = key
            items.append(item[0]+key+item[-1])
        else:
            items.append(item)
    newline = ''.join(items)
    items = []
    expr_keys = []
    # start of KGEN
    #if newline and newline.find('[')>=0: paren = '[]'
    #else: paren='()'
#    for item in splitparen(newline, paren='()'):
#        if isinstance(item, ParenString) and not _is_name(item[1:-1].strip()):
#            if item[1:-1].find('[') >= 0:
#                for item2 in splitparen(item, paren='[]'):
#                    if isinstance(item2, ParenString) and not _is_name(item2[1:-1].strip()):
#                        key = rev_string_map.get(item2)
#                        if key is None:
#                            _cache['pindex'] += 1
#                            index = _cache['pindex']
#                            key = 'F2PY_EXPR_TUPLE_%s' % (index)
#                            it = item2[1:-1].strip()
#                            string_map[key] = it
#                            rev_string_map[it] = key
#                            expr_keys.append(key)
#                        items.append(item2[0]+key+item2[-1])
#                    else:
#                        items.append(item2)
#            else:
#                key = rev_string_map.get(item)
#                if key is None:
#                    _cache['pindex'] += 1
#                    index = _cache['pindex']
#                    key = 'F2PY_EXPR_TUPLE_%s' % (index)
#                    it = item[1:-1].strip()
#                    string_map[key] = it
#                    rev_string_map[it] = key
#                    expr_keys.append(key)
#                items.append(item[0]+key+item[-1])
#        else:
#            items.append(item)
    # end of KGEN addition
    # start of KGEN deletion
    for item in splitparen(newline):
        if isinstance(item, ParenString) and not _is_name(item[1:-1].strip()):
            key = rev_string_map.get(item)
            if key is None:
                _cache['pindex'] += 1
                index = _cache['pindex']
                key = 'F2PY_EXPR_TUPLE_%s' % (index)
                it = item[1:-1].strip()
                string_map[key] = it
                rev_string_map[it] = key
                expr_keys.append(key)
            items.append(item[0]+key+item[-1])
        else:
             items.append(item)
    # end of KGEN deletion

    found_keys = set()
    for k in expr_keys:
        v = string_map[k]
        l = _f2py_str_findall(v)
        if l:
            found_keys = found_keys.union(l)
            for k1 in l:
                v = v.replace(k1, string_map[k1])
            string_map[k] = v
    for k in found_keys:
        del string_map[k]
    return ''.join(items), string_map

def splitquote(line, stopchar=None, lower=False, quotechars = '"\''):
    """
    Fast LineSplitter
    """
    items = []
    i = 0
    while 1:
        try:
            char = line[i]; i += 1
        except IndexError:
            break
        l = []
        l_append = l.append
        nofslashes = 0
        if stopchar is None:
            # search for string start
            while 1:
                if char in quotechars and not nofslashes % 2:
                    stopchar = char
                    i -= 1
                    break
                if char=='\\':
                    nofslashes += 1
                else:
                    nofslashes = 0
                l_append(char)
                try:
                    char = line[i]; i += 1
                except IndexError:
                    break
            if not l: continue
            item = ''.join(l)
            if lower: item = item.lower()
            items.append(item)
            continue
        if char==stopchar:
            # string starts with quotechar
            l_append(char)
            try:
                char = line[i]; i += 1
            except IndexError:
                if l:
                    item = String(''.join(l))
                    items.append(item)
                break
        # else continued string
        while 1:
            if char==stopchar and not nofslashes % 2:
                l_append(char)
                stopchar = None
                break
            if char=='\\':
                nofslashes += 1
            else:
                nofslashes = 0
            l_append(char)
            try:
                char = line[i]; i += 1
            except IndexError:
                break
        if l:
            item = String(''.join(l))
            items.append(item)
    return items, stopchar

class LineSplitterBase(object):

    def __iter__(self):
        return self

    def next(self):
        item = ''
        while not item:
            item = self.get_item() # get_item raises StopIteration
        return item

class LineSplitter(LineSplitterBase):
    """ Splits a line into non strings and strings. E.g.
    abc=\"123\" -> ['abc=','\"123\"']
    Handles splitting lines with incomplete string blocks.
    """
    def __init__(self, line,
                 quotechar = None,
                 lower=False,
                 ):
        self.fifo_line = [c for c in line]
        self.fifo_line.reverse()
        self.quotechar = quotechar
        self.lower = lower

    def split2(self):
        """
        Split line until the first start of a string.
        """
        try:
            item1 = self.get_item()
        except StopIteration:
            return '',''
        i = len(item1)
        l = self.fifo_line[:]
        l.reverse()
        item2 = ''.join(l)
        return item1,item2

    def get_item(self):
        fifo_pop = self.fifo_line.pop
        try:
            char = fifo_pop()
        except IndexError:
            raise StopIteration
        fifo_append = self.fifo_line.append
        quotechar = self.quotechar
        l = []
        l_append = l.append

        nofslashes = 0
        if quotechar is None:
            # search for string start
            while 1:
                if char in '"\'' and not nofslashes % 2:
                    self.quotechar = char
                    fifo_append(char)
                    break
                if char=='\\':
                    nofslashes += 1
                else:
                    nofslashes = 0
                l_append(char)
                try:
                    char = fifo_pop()
                except IndexError:
                    break
            item = ''.join(l)
            if self.lower: item = item.lower()
            return item

        if char==quotechar:
            # string starts with quotechar
            l_append(char)
            try:
                char = fifo_pop()
            except IndexError:
                return String(''.join(l))
        # else continued string
        while 1:
            if char==quotechar and not nofslashes % 2:
                l_append(char)
                self.quotechar = None
                break
            if char=='\\':
                nofslashes += 1
            else:
                nofslashes = 0
            l_append(char)
            try:
                char = fifo_pop()
            except IndexError:
                break
        return String(''.join(l))

def splitparen(line,paren='()'):
    """
    Fast LineSplitterParen.
    """
    stopchar = None
    startchar, endchar = paren[0],paren[1]

    items = []
    i = 0
    while 1:
        try:
            char = line[i]; i += 1
        except IndexError:
            break
        nofslashes = 0
        l = []
        l_append = l.append
        if stopchar is None:
            # search for parenthesis start
            while 1:
                if char==startchar and not nofslashes % 2:
                    stopchar = endchar
                    i -= 1
                    break
                if char=='\\':
                    nofslashes += 1
                else:
                    nofslashes = 0
                l_append(char)
                try:
                    char = line[i]; i += 1
                except IndexError:
                    break
            item = ''.join(l)
        else:
            nofstarts = 0
            while 1:
                if char==stopchar and not nofslashes % 2 and nofstarts==1:
                    l_append(char)
                    stopchar = None
                    break
                if char=='\\':
                    nofslashes += 1
                else:
                    nofslashes = 0
                if char==startchar:
                    nofstarts += 1
                elif char==endchar:
                    nofstarts -= 1
                l_append(char)
                try:
                    char = line[i]; i += 1
                except IndexError:
                    break
            item = ParenString(''.join(l))
        items.append(item)
    return items

class LineSplitterParen(LineSplitterBase):
    """ Splits a line into strings and strings with parenthesis. E.g.
    a(x) = b(c,d) -> ['a','(x)',' = b','(c,d)']
    """
    def __init__(self, line, paren = '()'):
        self.fifo_line = [c for c in line]
        self.fifo_line.reverse()
        self.startchar = paren[0]
        self.endchar = paren[1]
        self.stopchar = None

    def get_item(self):
        fifo_pop = self.fifo_line.pop
        try:
            char = fifo_pop()
        except IndexError:
            raise StopIteration
        fifo_append = self.fifo_line.append
        startchar = self.startchar
        endchar = self.endchar
        stopchar = self.stopchar
        l = []
        l_append = l.append

        nofslashes = 0
        if stopchar is None:
            # search for parenthesis start
            while 1:
                if char==startchar and not nofslashes % 2:
                    self.stopchar = endchar
                    fifo_append(char)
                    break
                if char=='\\':
                    nofslashes += 1
                else:
                    nofslashes = 0
                l_append(char)
                try:
                    char = fifo_pop()
                except IndexError:
                    break
            item = ''.join(l)
            return item

        nofstarts = 0
        while 1:
            if char==stopchar and not nofslashes % 2 and nofstarts==1:
                l_append(char)
                self.stopchar = None
                break
            if char=='\\':
                nofslashes += 1
            else:
                nofslashes = 0
            if char==startchar:
                nofstarts += 1
            elif char==endchar:
                nofstarts -= 1
            l_append(char)
            try:
                char = fifo_pop()
            except IndexError:
                break
        return ParenString(''.join(l))

def test():
    splitter = LineSplitter('abc\\\' def"12\\"3""56"dfad\'a d\'')
    l = [item for item in splitter]
    assert l==['abc\\\' def','"12\\"3"','"56"','dfad','\'a d\''],str(l)
    assert splitter.quotechar is None
    l,stopchar=splitquote('abc\\\' def"12\\"3""56"dfad\'a d\'')
    assert l==['abc\\\' def','"12\\"3"','"56"','dfad','\'a d\''],str(l)
    assert stopchar is None

    splitter = LineSplitter('"abc123&')
    l = [item for item in splitter]
    assert l==['"abc123&'],str(l)
    assert splitter.quotechar=='"'
    l,stopchar = splitquote('"abc123&')
    assert l==['"abc123&'],str(l)
    assert stopchar=='"'

    splitter = LineSplitter(' &abc"123','"')
    l = [item for item in splitter]
    assert l==[' &abc"','123']
    assert splitter.quotechar is None
    l,stopchar = splitquote(' &abc"123','"')
    assert l==[' &abc"','123']
    assert stopchar is None

    l = split2('')
    assert l==('',''),str(l)
    l = split2('12')
    assert l==('12',''),str(l)
    l = split2('1"a"//"b"')
    assert l==('1','"a"//"b"'),str(l)
    l = split2('"ab"')
    assert l==('','"ab"'),str(l)

    splitter = LineSplitterParen('a(b) = b(x,y(1)) b\((a)\)')
    l = [item for item in splitter]
    assert l==['a', '(b)', ' = b', '(x,y(1))', ' b\\(', '(a)', '\\)'],str(l)
    l = splitparen('a(b) = b(x,y(1)) b\((a)\)')
    assert l==['a', '(b)', ' = b', '(x,y(1))', ' b\\(', '(a)', '\\)'],str(l)

    l = string_replace_map('a()')
    print(l)
    print('ok')

if __name__ == '__main__':
    test()



import os, io, locale, math, random

from collections import OrderedDict
from microapp import App
from fortlab.kggenfile import (
    genkobj,
    gensobj,
    KERNEL_ID_0,
    event_register,
    create_rootnode,
    create_programnode,
    init_plugins,
    append_program_in_root,
    set_indent,
    plugin_config,
)
from fortlab.kgutils import (
    ProgramException,
    UserException,
    remove_multiblanklines,
    run_shcmd,
    tounicode,
)
from fortlab.resolver.kgparse import KGGenType
from fortlab.kgextra import (
    kgen_utils_file_head,
    kgen_utils_file_checksubr,
    kgen_get_newunit,
    kgen_error_stop,
    kgen_utils_file_tostr,
    kgen_utils_array_sumcheck,
    kgen_rankthread,
)

here = os.path.abspath(os.path.realpath(os.path.dirname(__file__)))
KGUTIL = "kgen_utils.f90"


class FortranVariableAnalyzer(App):
    _name_ = "vargen"
    _version_ = "0.1.0"

    def __init__(self, mgr):
        self.add_argument("analysis", help="analysis object")
        self.add_argument("--outdir", help="output directory")

        self.register_forward("kerneldir", help="kernel generation code directory")

    def perform(self, args):

        self.config = args.analysis["_"]

        self.config['modelfile'] = 'model.ini'
        self.config['model'] = OrderedDict()
        self.config['model']['path'] = ""
        self.config['model']['reuse_rawdata'] = True
        self.config['model']['types'] = OrderedDict()
        self.config['model']['types']['code'] = OrderedDict()
        self.config['model']['types']['code']['id'] = '0'
        self.config['model']['types']['code']['name'] = 'code'
        self.config['model']['types']['code']['collector'] = 'codecollect'
        self.config['model']['types']['code']['combiner'] = 'codecombine'
        self.config['model']['types']['code']['percentage'] = 99.9
        self.config['model']['types']['code']['filter'] = None
        self.config['model']['types']['code']['ndata'] = 20
        self.config['model']['types']['code']['enabled'] = False
        self.config['model']['types']['etime'] = OrderedDict()
        self.config['model']['types']['etime']['id'] = '1'
        self.config['model']['types']['etime']['name'] = 'etime'
        self.config['model']['types']['etime']['collector'] = 'timingcollect'
        self.config['model']['types']['etime']['combiner'] = 'timingcombine'
        self.config['model']['types']['etime']['nbins'] = 5
        self.config['model']['types']['etime']['ndata'] = 20
        self.config['model']['types']['etime']['minval'] = None
        self.config['model']['types']['etime']['maxval'] = None
        self.config['model']['types']['etime']['timer'] = None
        self.config['model']['types']['etime']['enabled'] = True
        self.config['model']['types']['papi'] = OrderedDict()
        self.config['model']['types']['papi']['id'] = '2'
        self.config['model']['types']['papi']['name'] = 'papi'
        self.config['model']['types']['papi']['collector'] = 'papicollect'
        self.config['model']['types']['papi']['combiner'] = 'papicombine'
        self.config['model']['types']['papi']['nbins'] = 5
        self.config['model']['types']['papi']['ndata'] = 20
        self.config['model']['types']['papi']['minval'] = None
        self.config['model']['types']['papi']['maxval'] = None
        self.config['model']['types']['papi']['header'] = None
        self.config['model']['types']['papi']['event'] = 'PAPI_TOT_INS'
        self.config['model']['types']['papi']['static'] = None
        self.config['model']['types']['papi']['dynamic'] = None
        self.config['model']['types']['papi']['enabled'] = False
        args.outdir = args.outdir["_"] if args.outdir else os.getcwd()

        if not os.path.exists(args.outdir):
            os.makedirs(args.outdir)

        self._trees = []
        self.genfiles = []

        self.config["used_srcfiles"].clear()

        state_realpath = os.path.realpath(os.path.join(args.outdir, "state"))
        kernel_realpath = os.path.realpath(os.path.join(args.outdir, "kernel"))

        self.config["path"]["kernel_output"] = kernel_realpath
        self.config["path"]["state_output"] = state_realpath

        self.add_forward(kerneldir=kernel_realpath)

        if not os.path.exists(kernel_realpath):
            os.makedirs(kernel_realpath)

        gencore_plugindir = os.path.join(here, "plugins", "gencore")
        varlist_plugindir = os.path.join(here, "plugins", "genvarlist")

        plugins = (
            ("ext.gencore", gencore_plugindir),
            ("ext.genvarlist", varlist_plugindir),
        )

        init_plugins([KERNEL_ID_0], plugins, self.config)
        plugin_config["current"].update(self.config)

        for filepath, (srcobj, mods_used, units_used) in self.config[
            "srcfiles"].items():
            if hasattr(srcobj.tree, "geninfo") and KGGenType.has_state(
                srcobj.tree.geninfo
            ):
                kfile = genkobj(None, srcobj.tree, KERNEL_ID_0)
                sfile = gensobj(None, srcobj.tree, KERNEL_ID_0)
                sfile.kgen_stmt.used4genstate = False
                if kfile is None or sfile is None:
                    raise kgutils.ProgramException(
                        "Kernel source file is not generated for %s." % filepath
                    )
                self.genfiles.append((kfile, sfile, filepath))
                self.config["used_srcfiles"][filepath] = (
                    kfile,
                    sfile,
                    mods_used,
                    units_used,
                )

        for plugin_name in event_register.keys():
            if not plugin_name.startswith("ext"):
                continue
            for kfile, sfile, filepath in self.genfiles:
                kfile.created([plugin_name])
                sfile.created([plugin_name])

            for tree in self._trees:
                tree.created([plugin_name])

        for plugin_name in event_register.keys():
            if not plugin_name.startswith("ext"):
                continue
            for kfile, sfile, filepath in self.genfiles:
                kfile.process([plugin_name])
                sfile.process([plugin_name])

            for tree in self._trees:
                tree.process([plugin_name])

        for plugin_name in event_register.keys():
            if not plugin_name.startswith("ext"):
                continue
            for kfile, sfile, filepath in self.genfiles:
                kfile.finalize([plugin_name])
                sfile.finalize([plugin_name])

            for tree in self._trees:
                tree.finalize([plugin_name])

        for plugin_name in event_register.keys():
            if not plugin_name.startswith("ext"):
                continue
            for kfile, sfile, filepath in self.genfiles:
                kfile.flatten(KERNEL_ID_0, [plugin_name])
                sfile.flatten(KERNEL_ID_0, [plugin_name])

            for tree in self._trees:
                tree.flatten(KERNEL_ID_0, [plugin_name])

        kernel_files = []
        state_files = []
        enc = locale.getpreferredencoding(False)

        for kfile, sfile, filepath in self.genfiles:
            filename = os.path.basename(filepath)
            set_indent("")
            klines = kfile.tostring()
            if klines is not None:
                klines = remove_multiblanklines(klines)
                kernel_files.append(filename)
                with io.open(
                    os.path.join(kernel_realpath, filename), "w", encoding=enc
                ) as (fd):
                    fd.write(tounicode(klines))

        if self.config["state_switch"]["clean"]:
            run_shcmd(self.config["state_switch"]["clean"])

        return



# simple_timing.py

#import os 
from fortlab.resolver import statements, block_statements, typedecl_statements
from fortlab.resolver.block_statements import SubProgramStatement, Function, Subroutine, Module
from fortlab.resolver.kgparse import KGGenType
from fortlab.kgplugin import Kgen_Plugin

class GenVar(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

    # registration
    def register(self, msg):
        self.frame_msg = msg

        self.callsite_stmts = getinfo('callsite_stmts')

        # register initial events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.FINISH_PROCESS, \
            self.callsite_stmts[0], None, self.create_callsite_varlist)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.FINISH_PROCESS, \
            SubProgramStatement, None, self.create_subp_varlist)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.FINISH_PROCESS, \
            Module, None, self.create_module_varlist)

    def create_module_varlist(self, node):

        for typedecl in node.kgen_stmt.typedecl_resolvers:
            for uname, req in KGGenType.get_state_in(typedecl.geninfo):
                # add comment at referrer
                obj = req.originator.genkpair
                idx, name, part = get_part_index(obj)
                line = str(typedecl.item.span)
                cmt = "\"%s\" is defined at module \"%s\" near original line %s" % (uname.last(), node.name, line)
                part_insert_comment(obj.kgen_parent, name, idx, cmt)

                # add comment at referee
                obj = typedecl.genkpair
                idx, name, part = get_part_index(obj)
                ref = ":".join(uname.namelist[:-1])
                line = str(uname.stmt.item.span)
                cmt = "\"%s\" is referenced from namepath of \"%s\" near original line %s" % (uname.last(), ref, line)
                part_insert_comment(obj.kgen_parent, name, idx, cmt)

    def create_subp_varlist(self, node):

        idx, name, part = get_part_index(node)

        part_insert_comment(node.kgen_parent, name, idx, "")

        if hasattr(node.kgen_stmt, "geninfo"):
            for gendata in node.kgen_stmt.geninfo.values():
                for uname, res in gendata:
                    call = ":".join(uname.namelist[:-1])
                    line = str(uname.stmt.item.span)
                    cmt = "referenced from \"%s\" near original line %s" % (call, line)
                    part_insert_comment(node.kgen_parent, name, idx, cmt)

                    call = ":".join(uname.namelist[:-1])
                    line = str(uname.stmt.item.span)
                    cmt = "referenced from \"%s\" near original line %s" % (call, line)
                    part_insert_comment(node.kgen_parent, name, idx, cmt)

        part_insert_comment(node.kgen_parent, name, idx, "")
        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "This subprogram is called from following callsites:")
        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "")


        # TODO: mark at originator and destination too
        #for typedecl in node.kgen_stmt.typedecl_resolvers:
        #    for uname, req in KGGenType.get_state_in(typedecl.geninfo):
        #        import pdb;pdb.set_trace()


    def create_callsite_varlist(self, node):
  
        idx, name, part = get_part_index(node)

        # VAR

        for insubr in getinfo("modreadsubrs").values():
            for stmt in get_part(insubr, EXEC_PART):
                #part_append_node(node.kgen_parent, name, stmt)
                part_insert_node(node.kgen_parent, name, idx, stmt)

        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "External variables possibly used for calculating value(s) for other variable(s)")
        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "")

        for outsubr in getinfo("modwritesubrs").values():
            for stmt in get_part(outsubr, EXEC_PART):
                #part_append_node(node.kgen_parent, name, stmt)
                part_insert_node(node.kgen_parent, name, idx, stmt)

        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "External variables possibly being modified in the following part")
        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "")

        for stmt in getinfo("localread"):
            part_insert_node(node.kgen_parent, name, idx, stmt)

        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "Local variables possibly used for calculating value(s) for other variable(s)")
        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "")

        for stmt in getinfo("localwrite"):
            part_insert_node(node.kgen_parent, name, idx, stmt)

        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "Local variables possibly being modified in the following part")
        part_insert_comment(node.kgen_parent, name, idx, "#########################################################")
        part_insert_comment(node.kgen_parent, name, idx, "")


#        # register event per function 
#        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.NODE_CREATED, \
#            getinfo('parentblock_stmt'), None, self.register_event) 
#
#    def register_event(self, node):
#
#        attrs = {'type_spec': 'INTEGER', 'selector': ('8', None), \
#            'entity_decls': ['kgen_start_clock', 'kgen_stop_clock', 'kgen_rate_clock']}
#        part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs) 
#
#        attrs = {'type_spec': 'REAL', 'selector': (None, 'kgen_dp'), 'entity_decls': ['gkgen_measure']}
#        part_append_genknode(node, DECL_PART, typedecl_statements.Real, attrs=attrs) 
#
#        prenode = getinfo('blocknode_aftercallsite_main')
#        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
#            prenode, None, self.add_execblock)
#
#    def add_execblock(self, node):
#
#        attrs = {'designator': 'SYSTEM_CLOCK', 'items': ['kgen_start_clock', 'kgen_rate_clock']}
#        part_append_genknode(node, EXEC_PART, statements.Call, attrs=attrs)
#       
#        attrs = {'loopcontrol': 'kgen_intvar = 1, KGEN_MAXITER'}
#        doiter = part_append_genknode(node, EXEC_PART, block_statements.Do, attrs=attrs)
#
#        kernel_stmts = getinfo('callsite_stmts')
#        start = kernel_stmts[0].item.span[0]-1
#        end = kernel_stmts[-1].item.span[1]
#        lines = kernel_stmts[0].top.prep[start:end]
#        lines_str = '\n'.join(lines)
#        dummy_node = part_append_genknode(doiter, EXEC_PART, statements.Call)
#        dummy_node.kgen_stmt = getinfo('dummy_stmt')
#        dummy_node.kgen_forced_line = lines_str
#
#        attrs = {'designator': 'SYSTEM_CLOCK', 'items': ['kgen_stop_clock', 'kgen_rate_clock']}
#        part_append_genknode(node, EXEC_PART, statements.Call, attrs=attrs)
#
#        if getinfo('cache_pollution'):
#            attrs = {'expr': 'kgen_cache_control(1) == 0'}
#            ifcache = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)
#
#            attrs = {'variable': 'kgen_stop_clock', 'sign': '=', 'expr': 'kgen_stop_clock  + kgen_cache_control(1)'}
#            part_append_genknode(ifcache, EXEC_PART, statements.Assignment, attrs=attrs)
#
#        attrs = {'variable': 'kgen_measure', 'sign': '=', 'expr': '1.0D6*(kgen_stop_clock - kgen_start_clock)/DBLE(kgen_rate_clock*KGEN_MAXITER)'}
#        part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)
#
#        if getinfo('add_mpi_frame'):
#            part_append_comment(node, EXEC_PART, '#ifdef _MPI', style='rawtext')
#            part_append_comment(node, EXEC_PART, 'CALL mpi_allreduce(kgen_measure, gkgen_measure, 1, mpi_real8, mpi_max, mpi_comm_world, kgen_ierr)', style='rawtext')
#            attrs = {'variable': 'kgen_measure', 'sign': '=', 'expr': 'gkgen_measure'}
#            part_append_genknode(node, EXEC_PART, statements.Assignment, attrs=attrs)
#            part_append_comment(node, EXEC_PART, '#endif', style='rawtext')
#
#        attrs = {'expr': 'check_status%rank==0'}
#        ifrank = part_append_genknode(node, EXEC_PART, block_statements.IfThen, attrs=attrs)
#
#        attrs = {'items': ['"%s : Time per call (usec): "'%getinfo('kernel_name'), 'kgen_measure']}
#        part_append_gensnode(ifrank, EXEC_PART, statements.Write, attrs=attrs)




import os, time, glob, shutil, locale, io

from collections import OrderedDict
from configparser import ConfigParser
from microapp import App

from fortlab.kgutils import logger, remove_multiblanklines, dequote, UserException
from fortlab.kggenfile import gensobj, KERNEL_ID_0, event_register, Gen_Statement, set_indent, init_plugins


class FortranTimingGenerator(App):

    _name_ = "timinggen"
    _version_ = "0.1.0"

    def __init__(self, mgr):

        self.config = None
        self.genfiles = []

        self.add_argument("analysis", help="analysis object")
        self.add_argument("--cleancmd", type=str, help="Software clean command.")
        self.add_argument("--buildcmd", metavar="build command", help="Software build command")
        self.add_argument("--runcmd", metavar="run command", help="Software run command")
        self.add_argument("--outdir", help="output directory")
        self.add_argument("--no-cache", action="store_true",
                            help="force to collect timing data")

        self.register_forward("etimedir", help="elapsedtime instrumented code directory")
        self.register_forward("modeldir", help="elapsedtime data directory")

    def perform(self, args):

        self.config = args.analysis["_"]

        if args.cleancmd:
            self.config["cmd_clean"]['cmds'] = dequote(args.cleancmd["_"])

        if args.buildcmd:
            self.config["cmd_build"]['cmds'] = dequote(args.buildcmd["_"])

        else:
            raise UserException("'--buildcmd' option is not given.")

        if args.runcmd:
            self.config["cmd_run"]['cmds'] = dequote(args.runcmd["_"])

        else:
            raise UserException("'--runcmd' option is not given.")

        # create directory if needed
        args.outdir = args.outdir["_"] if args.outdir else os.getcwd()

        if not os.path.exists(args.outdir):
            os.makedirs(args.outdir)

        srcfiles = OrderedDict()

        #etimedir = os.path.join(args.outdir, "etime")

        model_realpath = os.path.realpath(os.path.join(args.outdir, "model"))
        etime_realpath = os.path.realpath(os.path.join(args.outdir, "etime"))

        self.add_forward(etimedir=etime_realpath, modeldir=model_realpath)

        data_etime_path = os.path.join(model_realpath, "__data__",
                            self.config["model"]['types']['etime']['id'])

        if not os.path.exists(etime_realpath):
            os.makedirs(etime_realpath)

        if (os.path.exists(data_etime_path) and len(glob.glob(os.path.join(
                data_etime_path, "*"))) > 0 and
                self.config["model"]['reuse_rawdata']):
            logger.info('Reusing elapsedtime raw data.')

        else:
            logger.info('Generating elapsedtime raw data.')

            if os.path.exists(data_etime_path):
                shutil.rmtree(data_etime_path)

            rsc_etime_path = os.path.join(model_realpath, "__data__",
                "__resource__", self.config["model"]['types']['etime']['id'])

            if os.path.exists(rsc_etime_path):
                shutil.rmtree(rsc_etime_path)

            time.sleep(1)

            os.makedirs(data_etime_path)
            os.makedirs(rsc_etime_path)

            # generate wrapper nodes
            for filepath, (srcobj, mods_used, units_used) in self.config["srcfiles"].items():
                if os.path.realpath(filepath) == os.path.realpath(self.config["callsite"]['filepath']):
                    sfile = gensobj(None, srcobj.tree, KERNEL_ID_0)
                    sfile.used4etime = True
                    self.genfiles.append((sfile, filepath))
                    self.config["used_srcfiles"][filepath] = (sfile, mods_used, units_used)


#                here = os.path.dirname(os.path.realpath(__file__))
#                etime_plugindir = os.path.join(here, "plugins", "gencore")
#
#                self.config["path"]["model_path"] = model_realpath 
#
#                init_plugins([KERNEL_ID_0], {'etime.gencore': etime_plugindir}, self.config)

            # process each nodes in the tree
            for plugin_name in event_register.keys():
                if not plugin_name.startswith('etime'): continue

                for sfile, filepath in self.genfiles:
                    sfile.created([plugin_name])

            for plugin_name in event_register.keys():
                if not plugin_name.startswith('etime'): continue

                for sfile, filepath in self.genfiles:
                    sfile.process([plugin_name])

            for plugin_name in event_register.keys():
                if not plugin_name.startswith('etime'): continue

                for sfile, filepath in self.genfiles:
                    sfile.finalize([plugin_name])

            for plugin_name in event_register.keys():
                if not plugin_name.startswith('etime'): continue

                for sfile, filepath in self.genfiles:
                    sfile.flatten(KERNEL_ID_0, [plugin_name])

            # generate source files from each node of the tree
            etime_files = []
            #import pdb; pdb.set_trace()
            for sfile, filepath in self.genfiles:
                #import pdb; pdb.set_trace()
                filename = os.path.basename(filepath)
                if sfile.used4etime:
                    set_indent('')
                    slines = sfile.tostring()
                    if slines is not None:
                        slines = remove_multiblanklines(slines)
                        etime_files.append(filename)
                        enc = locale.getpreferredencoding(False)
                        with io.open(os.path.join(etime_realpath, filename), 'w', encoding=enc) as fd:
                            fd.write(slines)
                        with io.open(os.path.join(etime_realpath, filename+".kgen"), 'w', encoding=enc) as ft:
                            ft.write('\n'.join(sfile.kgen_stmt.prep))

            self.gen_makefile(etime_realpath)

    def addsection(self, modeltype, section, options):

        modelfile = os.path.join(self.config["path"]["outdir"], self.config["modelfile"])

        mode = 'r+'
        if not os.path.exists(modelfile):
            raise Exception('Modelfile does not exists: %s'%modelfile)

        cfg = ConfigParser()
        cfg.optionxform = str
        cfg.read(modelfile)

        subsec = '%s.%s'%(modeltype, section)
        if cfg.has_section(subsec):
            raise Exception('Section already exists: %s'%subsec)

        cfg.add_section(subsec)

        for opt, val in options:
            cfg.set(subsec, opt, val)

        with io.open(modelfile, mode) as mf:
            cfg.write(mf)


    def addmodel(self, modeltype, sections):

        modelfile = '%s/%s'%(self.config["path"]['outdir'], self.config["modelfile"])

        mode = 'r+'
        if not os.path.exists(modelfile):
            mode = 'w+'

        with io.open(modelfile, mode) as mf:
            mf.seek(0, os.SEEK_END)
            size = mf.tell()
            if size == 0:
                mf.write('; KGen Model Data File\n')

        cfg = ConfigParser()
        cfg.optionxform = str
        cfg.read(modelfile)

        if not cfg.has_section(GEN):
            cfg.add_section(GEN)

        if not cfg.has_option(GEN, modeltype):
            cfg.set(GEN, modeltype, ', '.join(sections))
#
#        for sec in sections:
#            secname = '%s.%s'%(modeltype, sec)
#            if not cfg.has_section(secname):
#                cfg.add_section(secname)

        with io.open(modelfile, mode) as mf:
            cfg.write(mf)
                            

    def gen_makefile(self, etime_realpath):

        org_files = [ filepath for filepath, (sfile, mods_used, units_used) in self.config["used_srcfiles"].items() if sfile.used4etime ]

        if not self.config["topblock"]['stmt'].reader.id in org_files:
            org_files.append(self.config["topblock"]['filepath'])

        enc = locale.getpreferredencoding(False)
        with io.open('%s/Makefile'%etime_realpath, 'w', encoding=enc) as f:

            self.write(f, '# Makefile for KGEN-generated instrumentation')
            self.write(f, '')

            cwd = os.getcwd()

            prerun_build_str = ''
            if self.config["prerun"]['build']:
                self.write(f, 'PRERUN_BUILD := %s'%self.config["prerun"]['build'])
                prerun_build_str = '${PRERUN_BUILD}; '

            prerun_run_str = ''
            if self.config["prerun"]['run']:
                self.write(f, 'PRERUN_RUN := %s'%self.config["prerun"]['run'])
                prerun_run_str = '${PRERUN_RUN}; '

            self.write(f, '')
            if len(self.config["cmd_run"]['cmds']) > 0:
                self.write(f, 'run: build')
                self.write(f, '%s%s'%(prerun_run_str, self.config["cmd_run"]['cmds']), t=True)
            else:
                self.write(f, 'echo "No information is provided to run. Please specify run commands using \'state-run\' command line option"; exit -1', t=True)
            self.write(f, '')

            if len(self.config["cmd_build"]['cmds'])>0:
                self.write(f, 'build: %s'%self.config["state_switch"]['type'])
                self.write(f, '%s%s'%(prerun_build_str, self.config["cmd_build"]['cmds']), t=True)
                for org_file in org_files:
                    self.write(f, 'mv -f %(f)s.kgen_org %(f)s'%{'f':org_file}, t=True)
            else:
                self.write(f, 'echo "No information is provided to build. Please specify build commands using \'state-build\' command line option"; exit -1', t=True)
            self.write(f, '')

            self.write(f, '%s: save'%self.config["state_switch"]['type'])
            if self.config["state_switch"]['type']=='replace':
                for org_file in org_files:
                    basename = os.path.basename(org_file)
                    self.write(f, 'rm -f %s'%org_file, t=True)
                    self.write(f, 'cp -f %(f1)s %(f2)s'%{'f1':basename, 'f2':org_file}, t=True)
            elif self.config["state_switch"]['type']=='copy':
                for org_file in org_files:
                    basename = os.path.basename(org_file)
                    self.write(f, 'rm -f %s/%s'%(self.config["state_switch"]['directory'], basename), t=True)
                    self.write(f, 'cp -f %s/%s %s'%(etime_realpath, basename, self.config["state_switch"]['directory']), t=True)
            self.write(f, '')

            self.write(f, 'recover:')
            for org_file in org_files:
                self.write(f, 'rm -f %s'%org_file, t=True)
                self.write(f, 'cp -f %s.kgen_org %s'%(os.path.basename(org_file), org_file), t=True)
            self.write(f, '')

            self.write(f, 'recover_from_srcdir:')
            for org_file in org_files:
                self.write(f, 'rm -f %s'%org_file, t=True)
                self.write(f, 'cp -f %(f)s.kgen_org %(f)s'%{'f':org_file}, t=True)
            self.write(f, '')


            self.write(f, 'save:')
            for org_file in org_files:
                self.write(f, 'if [ ! -f %(f)s.kgen_org ]; then cp -f %(f)s %(f)s.kgen_org; fi'%{'f':org_file}, t=True)
                self.write(f, 'if [ ! -f %(g)s.kgen_org ]; then cp -f %(f)s %(g)s.kgen_org; fi'%{'f':org_file, 'g':os.path.basename(org_file)}, t=True)
            self.write(f, '')

            if len(self.config["cmd_clean"]['cmds']) > 0:
                self.write(f, 'clean:')
                self.write(f, self.config["cmd_clean"]['cmds'], t=True)
            self.write(f, '')

    def write(self, f, line, n=True, t=False):
        nl = ''
        tab = ''
        if n: nl = '\n'
        if t: tab = '\t'
        text = tab + line + nl

        if type(text) == type(u""):
            f.write(text)

        else:
            enc = locale.getpreferredencoding(False)
            f.write(text.decode(enc))




import os
import random
import json
from fortlab.kgplugin import Kgen_Plugin
from fortlab.resolver import block_statements, statements, typedecl_statements
import collections

BEFORE_CALLSITE = 'before_callsite'
AFTER_CALLSITE = 'after_callsite'

class Gen_ElapsedTime_File(Kgen_Plugin):

    def __init__(self):
        self.frame_msg = None
        self.paths = collections.OrderedDict()
        self.logger = getinfo('logger')

    # registration
    def register(self, msg):

        self.frame_msg = msg

        # when begin process
        #self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
        #    getinfo('topblock_stmt'), None, self.save_maps)

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.STATE, GENERATION_STAGE.NODE_CREATED, \
            getinfo('parentblock_stmt'), None, self.save_etime)

    ##################################
    # printing paths
    ##################################

#    def save_maps(self, node):
#
#        # generate metadata.json for coverage
#        if os.path.exists('%s/__data__/modeltypes'%getinfo('model_path')):
#            json_data = None
#            with open('%s/__data__/modeltypes'%getinfo('model_path'), 'r') as cm:
#                json_data = json.load(cm)
#                if u'"%s"'%getinfo('etime_typeid') not in json_data[u'datamap']:
#                    json_data[u'datamap'][u'"%s"'%getinfo('etime_typeid')] = u'"%s"'%getinfo('etime_typename')
#
#            with open('%s/__data__/modeltypes'%getinfo('model_path'), 'w') as cm:
#                cm.write(json.dumps(json_data))
#        else:
#            with open('%s/__data__/modeltypes'%getinfo('model_path'), 'w') as fm:
#                fm.write(u'{"datatype": "model", "datamap": { "%s": "%s" }}\n'%\
#                    (getinfo('etime_typeid'), getinfo('etime_typename')))
#

    def has_mpifh(self, stmt):
        retval = False
        if hasattr(stmt, 'a') and hasattr(stmt.a, 'variable_names') and 'mpi_wtime' in stmt.a.variable_names:
            retval = True
        elif hasattr(stmt, 'parent'):
            retval = self.has_mpifh(stmt.parent)
        return retval

    def insert_timing_stmts(self, timer, kernel_id, part_before, part_after, isopenmp=False):

        if timer == 'mpiwtime':
            # start timer before callsite
            if isopenmp:
                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 0)', 'sign': '=', 'expr': 'MPI_WTIME()'}
                namedpart_append_gensnode(kernel_id, part_before, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 1)', 'sign': '=', 'expr': 'MPI_WTIME()'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 2)', 'sign': '=', 'expr': 'MPI_WTICK()'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)
            else:
                attrs = {'variable': 'kgen_timer(0)', 'sign': '=', 'expr': 'MPI_WTIME()'}
                namedpart_append_gensnode(kernel_id, part_before, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(1)', 'sign': '=', 'expr': 'MPI_WTIME()'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(2)', 'sign': '=', 'expr': 'MPI_WTICK()'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

        elif timer == 'ompwtime':
            if isopenmp:
                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 0)', 'sign': '=', 'expr': 'OMP_GET_WTIME()'}
                namedpart_append_gensnode(kernel_id, part_before, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 1)', 'sign': '=', 'expr': 'OMP_GET_WTIME()'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 2)', 'sign': '=', 'expr': 'OMP_GET_WTICK()'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)
            else:
                attrs = {'variable': 'kgen_timer(0)', 'sign': '=', 'expr': 'OMP_GET_WTIME()'}
                namedpart_append_gensnode(kernel_id, part_before, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(1)', 'sign': '=', 'expr': 'OMP_GET_WTIME()'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(2)', 'sign': '=', 'expr': 'OMP_GET_WTICK()'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

        elif timer == 'cputime':
            if isopenmp:

                attrs = {'designator': 'CPU_TIME', 'items': ['kgen_start(OMP_GET_THREAD_NUM())']}
                namedpart_append_gensnode(kernel_id, part_before, statements.Call, attrs=attrs)

                attrs = {'designator': 'CPU_TIME', 'items': ['kgen_stop(OMP_GET_THREAD_NUM())']}
                namedpart_append_gensnode(kernel_id, part_after, statements.Call, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 0)', 'sign': '=', 'expr': 'kgen_start(OMP_GET_THREAD_NUM())'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 1)', 'sign': '=', 'expr': 'kgen_stop(OMP_GET_THREAD_NUM())'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 2)', 'sign': '=', 'expr': '1.0'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)
            else:

                attrs = {'designator': 'CPU_TIME', 'items': ['kgen_start']}
                namedpart_append_gensnode(kernel_id, part_before, statements.Call, attrs=attrs)

                attrs = {'designator': 'CPU_TIME', 'items': ['kgen_stop']}
                namedpart_append_gensnode(kernel_id, part_after, statements.Call, attrs=attrs)

                attrs = {'variable': 'kgen_timer(0)', 'sign': '=', 'expr': 'kgen_start'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(1)', 'sign': '=', 'expr': 'kgen_stop'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(2)', 'sign': '=', 'expr': '1.0'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)


        elif timer == 'sysclock':
            if isopenmp:

                attrs = {'designator': 'SYSTEM_CLOCK', 'items': ['kgen_start_clock(OMP_GET_THREAD_NUM())', 'kgen_rate_clock(OMP_GET_THREAD_NUM())']}
                namedpart_append_gensnode(kernel_id, part_before, statements.Call, attrs=attrs)

                attrs = {'designator': 'SYSTEM_CLOCK', 'items': ['kgen_stop_clock(OMP_GET_THREAD_NUM())', 'kgen_rate_clock(OMP_GET_THREAD_NUM())']}
                namedpart_append_gensnode(kernel_id, part_after, statements.Call, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 0)', 'sign': '=', 'expr': 'DBLE(kgen_start_clock(OMP_GET_THREAD_NUM())) / DBLE(kgen_rate_clock(OMP_GET_THREAD_NUM()))'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 1)', 'sign': '=', 'expr': 'DBLE(kgen_stop_clock(OMP_GET_THREAD_NUM())) / DBLE(kgen_rate_clock(OMP_GET_THREAD_NUM()))'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 2)', 'sign': '=', 'expr': '1.0 / DBLE(kgen_rate_clock(OMP_GET_THREAD_NUM()))'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)
            else:

                attrs = {'designator': 'SYSTEM_CLOCK', 'items': ['kgen_start_clock', 'kgen_rate_clock']}
                namedpart_append_gensnode(kernel_id, part_before, statements.Call, attrs=attrs)

                attrs = {'designator': 'SYSTEM_CLOCK', 'items': ['kgen_stop_clock', 'kgen_rate_clock']}
                namedpart_append_gensnode(kernel_id, part_after, statements.Call, attrs=attrs)

                attrs = {'variable': 'kgen_timer(0)', 'sign': '=', 'expr': 'DBLE(kgen_start_clock) / DBLE(kgen_rate_clock)'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(1)', 'sign': '=', 'expr': 'DBLE(kgen_stop_clock) / DBLE(kgen_rate_clock)'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

                attrs = {'variable': 'kgen_timer(2)', 'sign': '=', 'expr': '1.0 / DBLE(kgen_rate_clock)'}
                namedpart_append_gensnode(kernel_id, part_after, statements.Assignment, attrs=attrs)

        else:

            raise Exception('Not supported timer type: %s'%timer)



    def save_etime(self, node):

        # add type decls for timers

        datapath = '%s/__data__'%getinfo('model_path')
        etimepath = '%s/%s'%(datapath, getinfo('etime_typeid'))

        # mpi_wtime, mpi_wtick
        # omp_get_wtime, omp_get_wtick
        # system_clock

        attrs = {'type_spec': 'CHARACTER', 'selector':('4096', None), 'entity_decls': ['datapath']}
        part_append_gensnode(node, DECL_PART, typedecl_statements.Character, attrs=attrs)

        attrs = {'type_spec': 'INTEGER', 'entity_decls': ['ierror']}
        part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        if getinfo('is_mpi_app'):

            for mod_name, use_names in getinfo('mpi_use'):
                attrs = {'name':mod_name, 'isonly': True, 'items':use_names}
                part_append_gensnode(node, USE_PART, statements.Use, attrs=attrs)

            attrs = {'type_spec': 'LOGICAL', 'attrspec': [ 'SAVE' ], 'entity_decls': ['kgen_initialized = .FALSE.']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Logical, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['myrank']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            if getinfo('etime_timer') is None or getinfo('etime_timer') == 'mpiwtime':
                if not self.has_mpifh(node.kgen_stmt):
                    attrs = {'type_spec': 'REAL', 'selector': (None, '8'), 'entity_decls': ['MPI_WTIME', 'MPI_WTICK']}
                    part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        if getinfo('is_openmp_app'):
            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['OMP_GET_THREAD_NUM']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'REAL', 'selector': (None, '8'), \
                'attrspec': [ 'DIMENSION(0:%d, 0:2)'%(getinfo('openmp_maxthreads')-1) ], 'entity_decls': ['kgen_timer']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'SAVE', 'DIMENSION(0:%d)'%(getinfo('openmp_maxthreads')-1) ], 'entity_decls': ['kgen_invokes = 0']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'SAVE', 'DIMENSION(0:%d)'%(getinfo('openmp_maxthreads')-1) ], 'entity_decls': ['dataunit']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            if getinfo('etime_timer') is None or getinfo('etime_timer') == 'ompwtime':
                attrs = {'type_spec': 'REAL', 'selector': (None, '8'), 'entity_decls': ['OMP_GET_WTIME', 'OMP_GET_WTICK']}
                part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)
            elif getinfo('etime_timer') == 'cputime':
                attrs = {'type_spec': 'REAL', 'selector': ('8', None), 'attrspec': [ 'DIMENSION(0:%d)'%(getinfo('openmp_maxthreads')-1) ], \
                    'entity_decls': ['kgen_start', 'kgen_stop']}
                part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            elif getinfo('etime_timer') == 'sysclock':
                attrs = {'type_spec': 'INTEGER', 'selector': ('8', None), 'attrspec': [ 'DIMENSION(0:%d)'%(getinfo('openmp_maxthreads')-1) ], \
                    'entity_decls': ['kgen_start_clock', 'kgen_stop_clock', 'kgen_rate_clock']}
                part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        else:

            attrs = {'type_spec': 'REAL', 'selector': (None, '8'), \
                'attrspec': [ 'DIMENSION(0:2)' ], 'entity_decls': ['kgen_timer']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'attrspec': [ 'SAVE' ], 'entity_decls': [ 'kgen_invokes = 0']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            if getinfo('etime_timer') is None or getinfo('etime_timer') == 'sysclock':
                attrs = {'type_spec': 'INTEGER', 'selector': ('8', None), \
                    'entity_decls': ['kgen_start_clock', 'kgen_stop_clock', 'kgen_rate_clock']}
                part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)
            elif getinfo('etime_timer') == 'cputime':
                attrs = {'type_spec': 'REAL', 'selector': ('8', None), 'entity_decls': ['kgen_start', 'kgen_stop']}
                part_append_genknode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['dataunit']}
            part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        #attrs = {'type_spec': 'CHARACTER', 'selector':('6', None), 'entity_decls': ['numthreadsstr']}
        #part_append_gensnode(node, DECL_PART, typedecl_statements.Character, attrs=attrs)

        #attrs = {'type_spec': 'INTEGER', 'entity_decls': [ 'invokes', 'visits', 'intnum' ]}
        #part_append_gensnode(node, DECL_PART, typedecl_statements.Integer, attrs=attrs)

        #attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['istrue']}
        #part_append_gensnode(node, DECL_PART, typedecl_statements.Logical, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'selector':('10', None), 'entity_decls': ['rankstr']}
        part_append_gensnode(node, DECL_PART, typedecl_statements.Character, attrs=attrs)

        attrs = {'type_spec': 'CHARACTER', 'selector':('6', None), 'entity_decls': ['threadstr']}
        part_append_gensnode(node, DECL_PART, typedecl_statements.Character, attrs=attrs)


        # ????measure overhead of timing measurement

        stmts = getinfo('callsite_stmts')

        # TODO: fix location mismatch error
        # replace part_append_gensnode to namedpart_append_gensnode

        topobj = stmts[0].parent.genspair

        index, partname, part = get_part_index(stmts[0].genspair)
        namedpart_create_subpart(topobj, BEFORE_CALLSITE, EXEC_PART, index=index)

        # get the last stmt
        prevstmt = stmts[-1]
        while prevstmt.parent in stmts:
            prevstmt = prevstmt.parent

        index, partname, part = get_part_index(prevstmt.genspair)
        namedpart_create_subpart(topobj, AFTER_CALLSITE, EXEC_PART, index=index+1)
        
        body_block = AFTER_CALLSITE

        if getinfo('is_mpi_app'):

            attrs = {'designator': 'MPI_INITIALIZED', 'items': [ 'kgen_initialized', 'ierror' ]}
            #part_append_gensnode(topobj, EXEC_PART, statements.Call, attrs=attrs)
            namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Call, attrs=attrs)

            attrs = {'expr': 'kgen_initialized .AND. ( ierror .EQ. 0 )'}
            #topobj = part_append_gensnode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
            ifinit = namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, block_statements.IfThen, attrs=attrs)

            body_block = "_BDYBLK_"
            namedpart_link_part(ifinit, body_block, EXEC_PART)

        if getinfo('is_openmp_app'):
            #part_append_comment(topobj, EXEC_PART, 'CRITICAL (kgen_etime)', style='openmp')
            #namedpart_append_comment(node.kgen_kernel_id, AFTER_CALLSITE, 'CRITICAL (kgen_etime)', style='openmp')
            namedpart_append_comment(node.kgen_kernel_id, body_block, 'CRITICAL (kgen_etime)', style='openmp')

        if getinfo('is_mpi_app'):

            attrs = {'designator': 'MPI_COMM_RANK', 'items': [ getinfo('mpi_comm'), 'myrank', 'ierror' ]}
            #part_append_gensnode(topobj, EXEC_PART, statements.Call, attrs=attrs)
            #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Call, attrs=attrs)
            namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Call, attrs=attrs)

            attrs = {'specs': [ 'rankstr', '"(I10)"' ], 'items': [ 'myrank' ]}
            #part_append_gensnode(topobj, EXEC_PART, statements.Write, attrs=attrs)
            #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Write, attrs=attrs)
            namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Write, attrs=attrs)


            if getinfo('is_openmp_app'):

                attrs = {'variable': 'kgen_invokes(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': 'kgen_invokes(OMP_GET_THREAD_NUM()) + 1'}
                namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

                attrs = {'specs': [ 'threadstr', '"(I6)"' ], 'items': [ 'OMP_GET_THREAD_NUM()' ]}
                #part_append_gensnode(topobj, EXEC_PART, statements.Write, attrs=attrs)
                #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Write, attrs=attrs)
                namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Write, attrs=attrs)

                timer = getinfo('etime_timer')
                if timer is None:
                    #self.insert_timing_stmts('mpiwtime', node.kgen_kernel_id, BEFORE_CALLSITE, AFTER_CALLSITE, isopenmp=True)
                    self.insert_timing_stmts('mpiwtime', node.kgen_kernel_id, BEFORE_CALLSITE, body_block, isopenmp=True)
                else:
                    #self.insert_timing_stmts(timer, node.kgen_kernel_id, BEFORE_CALLSITE, AFTER_CALLSITE, isopenmp=True)
                    self.insert_timing_stmts(timer, node.kgen_kernel_id, BEFORE_CALLSITE, body_block, isopenmp=True)

            else:

                attrs = {'variable': 'kgen_invokes', 'sign': '=', 'expr': 'kgen_invokes + 1'}
                namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

                attrs = {'specs': [ 'threadstr', '"(I1)"' ], 'items': [ '0' ]}
                #part_append_gensnode(topobj, EXEC_PART, statements.Write, attrs=attrs)
                #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Write, attrs=attrs)
                namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Write, attrs=attrs)

                timer = getinfo('etime_timer')
                if timer is None:
                    #self.insert_timing_stmts('mpiwtime', node.kgen_kernel_id, BEFORE_CALLSITE, AFTER_CALLSITE, isopenmp=False)
                    self.insert_timing_stmts('mpiwtime', node.kgen_kernel_id, BEFORE_CALLSITE, body_block, isopenmp=False)
                else:
                    #self.insert_timing_stmts(timer, node.kgen_kernel_id, BEFORE_CALLSITE, AFTER_CALLSITE, isopenmp=False)
                    self.insert_timing_stmts(timer, node.kgen_kernel_id, BEFORE_CALLSITE, body_block, isopenmp=False)

#                # start timer before callsite
#                attrs = {'variable': 'kgen_timer(0)', 'sign': '=', 'expr': 'MPI_WTIME()'}
#                namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)
#
#                # stop timer after callsite
#                attrs = {'variable': 'kgen_timer(1)', 'sign': '=', 'expr': 'MPI_WTIME()'}
#                namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Assignment, attrs=attrs)
#
#                # stop timer after callsite
#                attrs = {'variable': 'kgen_timer(2)', 'sign': '=', 'expr': 'MPI_WTICK()'}
#                namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Assignment, attrs=attrs)
#

        else:

            attrs = {'specs': [ 'rankstr', '"(I1)"' ], 'items': [ '0' ]}
            #part_append_gensnode(topobj, EXEC_PART, statements.Write, attrs=attrs)
            #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Write, attrs=attrs)
            namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Write, attrs=attrs)

            if getinfo('is_openmp_app'):

                attrs = {'variable': 'kgen_invokes(OMP_GET_THREAD_NUM())', 'sign': '=', 'expr': 'kgen_invokes(OMP_GET_THREAD_NUM()) + 1'}
                namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

                attrs = {'specs': [ 'threadstr', '"(I6)"' ], 'items': [ 'OMP_GET_THREAD_NUM()' ]}
                #part_append_gensnode(topobj, EXEC_PART, statements.Write, attrs=attrs)
                #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Write, attrs=attrs)
                namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Write, attrs=attrs)

                timer = getinfo('etime_timer')
                if timer is None:
                    #self.insert_timing_stmts('ompwtime', node.kgen_kernel_id, BEFORE_CALLSITE, AFTER_CALLSITE, isopenmp=True)
                    self.insert_timing_stmts('ompwtime', node.kgen_kernel_id, BEFORE_CALLSITE, body_block, isopenmp=True)
                else:
                    #self.insert_timing_stmts(timer, node.kgen_kernel_id, BEFORE_CALLSITE, AFTER_CALLSITE, isopenmp=True)
                    self.insert_timing_stmts(timer, node.kgen_kernel_id, BEFORE_CALLSITE, body_block, isopenmp=True)
#
#                # start timer before callsite
#                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 0)', 'sign': '=', 'expr': 'OMP_GET_WTIME()'}
#                namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)
#
#                # stop timer after callsite
#                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 1)', 'sign': '=', 'expr': 'OMP_GET_WTIME()'}
#                namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Assignment, attrs=attrs)
#
#                # stop timer after callsite
#                attrs = {'variable': 'kgen_timer(OMP_GET_THREAD_NUM(), 2)', 'sign': '=', 'expr': 'OMP_GET_WTICK()'}
#                namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Assignment, attrs=attrs)


            else:

                attrs = {'variable': 'kgen_invokes', 'sign': '=', 'expr': 'kgen_invokes + 1'}
                namedpart_append_gensnode(node.kgen_kernel_id, BEFORE_CALLSITE, statements.Assignment, attrs=attrs)

                attrs = {'specs': [ 'threadstr', '"(I1)"' ], 'items': [ '0' ]}
                #part_append_gensnode(topobj, EXEC_PART, statements.Write, attrs=attrs)
                #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Write, attrs=attrs)
                namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Write, attrs=attrs)

                timer = getinfo('etime_timer')
                if timer is None:
                    #self.insert_timing_stmts('sysclock', node.kgen_kernel_id, BEFORE_CALLSITE, AFTER_CALLSITE, isopenmp=False)
                    self.insert_timing_stmts('sysclock', node.kgen_kernel_id, BEFORE_CALLSITE, body_block, isopenmp=False)
                else:
                    #self.insert_timing_stmts(timer, node.kgen_kernel_id, BEFORE_CALLSITE, AFTER_CALLSITE, isopenmp=False)
                    self.insert_timing_stmts(timer, node.kgen_kernel_id, BEFORE_CALLSITE, body_block, isopenmp=False)

        # append mpirank.ompthread: invocation, starttime, stoptime, resolution

        if getinfo('is_openmp_app'):

            attrs = {'specs': [ 'datapath', '*' ], 'items': [ '"%s/" // TRIM(ADJUSTL(rankstr)) // "." // TRIM(ADJUSTL(threadstr))'%etimepath ]}
            #part_append_gensnode(topobj, EXEC_PART, statements.Write, attrs=attrs)
            #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Write, attrs=attrs)
            namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Write, attrs=attrs)

            attrs = {'specs': ['NEWUNIT=dataunit(OMP_GET_THREAD_NUM())', 'FILE=TRIM(ADJUSTL(datapath))', 'STATUS="UNKNOWN"', 'ACTION="WRITE"', 'ACCESS="SEQUENTIAL"', 'POSITION="APPEND"', 'IOSTAT=ierror']}
            #part_append_gensnode(topobj, EXEC_PART, statements.Open, attrs=attrs)
            #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Open, attrs=attrs)
            namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Open, attrs=attrs)

            attrs = {'expr': 'ierror .EQ. 0'}
            #ifopen = part_append_gensnode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
            #ifopen = namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, block_statements.IfThen, attrs=attrs)
            ifopen = namedpart_append_gensnode(node.kgen_kernel_id, body_block, block_statements.IfThen, attrs=attrs)

            attrs = {'specs': ['UNIT=dataunit(OMP_GET_THREAD_NUM())', 'FMT="(I16,1X,ES34.16,1X,ES34.16,1X,ES34.16)"' ], \
                'items': [ 'kgen_invokes(OMP_GET_THREAD_NUM())', 'kgen_timer(OMP_GET_THREAD_NUM(), 0)', \
                'kgen_timer(OMP_GET_THREAD_NUM(), 1)', 'kgen_timer(OMP_GET_THREAD_NUM(), 2)' ]}
            part_append_gensnode(ifopen, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'specs': ['UNIT=dataunit(OMP_GET_THREAD_NUM())']}
            part_append_gensnode(ifopen, EXEC_PART, statements.Close, attrs=attrs)

            part_append_gensnode(ifopen, EXEC_PART, statements.Else)

            attrs = {'items': ['"FILE OPEN ERROR: "', 'TRIM(ADJUSTL(datapath))', 'ierror']}
            part_append_gensnode(ifopen, EXEC_PART, statements.Print, attrs=attrs)

            #part_append_comment(topobj, EXEC_PART, 'END CRITICAL (kgen_etime)', style='openmp')
            #namedpart_append_comment(node.kgen_kernel_id, AFTER_CALLSITE, 'END CRITICAL (kgen_etime)', style='openmp')
            namedpart_append_comment(node.kgen_kernel_id, body_block, 'END CRITICAL (kgen_etime)', style='openmp')

        else:

            attrs = {'specs': [ 'datapath', '*' ], 'items': [ '"%s/" // TRIM(ADJUSTL(rankstr)) // "." // TRIM(ADJUSTL(threadstr))'%etimepath ]}
            #part_append_gensnode(topobj, EXEC_PART, statements.Write, attrs=attrs)
            #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Write, attrs=attrs)
            namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Write, attrs=attrs)

            attrs = {'specs': ['NEWUNIT=dataunit', 'FILE=TRIM(ADJUSTL(datapath))', 'ACTION="WRITE"', 'ACCESS="SEQUENTIAL"', 'POSITION="APPEND"', 'IOSTAT=ierror']}
            #part_append_gensnode(topobj, EXEC_PART, statements.Open, attrs=attrs)
            #namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, statements.Open, attrs=attrs)
            namedpart_append_gensnode(node.kgen_kernel_id, body_block, statements.Open, attrs=attrs)

            attrs = {'expr': 'ierror .EQ. 0'}
            #ifopen = part_append_gensnode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
            #ifopen = namedpart_append_gensnode(node.kgen_kernel_id, AFTER_CALLSITE, block_statements.IfThen, attrs=attrs)
            ifopen = namedpart_append_gensnode(node.kgen_kernel_id, body_block, block_statements.IfThen, attrs=attrs)

            attrs = {'specs': ['UNIT=dataunit', 'FMT="(I16,1X,ES34.16,1X,ES34.16,1X,ES34.16)"' ], \
                'items': [ 'kgen_invokes', 'kgen_timer(0)', \
                'kgen_timer(1)', 'kgen_timer(2)' ]}
            part_append_gensnode(ifopen, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'specs': ['UNIT=dataunit']}
            part_append_gensnode(ifopen, EXEC_PART, statements.Close, attrs=attrs)

            part_append_gensnode(ifopen, EXEC_PART, statements.Else)

            attrs = {'items': ['"FILE OPEN ERROR: "', 'TRIM(ADJUSTL(datapath))', 'ierror']}
            part_append_gensnode(ifopen, EXEC_PART, statements.Print, attrs=attrs)




# gen_core_utils.py

# NOTES for supporting linked list
# STATE part
# A) save address of all saved derived type variables
# B) when a a pointer derived type is being saved, first check if the address of the target
#    variable is in the address list
#    a) if exits, do not save and leave a mark
#    b) if not exists, save and put the address of the saved variable in the address list
#
# KERNEL part
# A) when a pointer derived type is being read, first check if a mark exists,
#   a) If not exists, read and save the address of the derived type variable in dtype list with a mark
#   b) If exists, do not read and associate to a derived type variable in the dtype list according to the mark
#
# Mark
# A) derived type id
# B) IO id
#
# Address list
# A) an array of integers representing addresss
# B) generate in (named?) data block using common block
# C) external functions for using the address list
#
# Dtype list
# A) Two separate one dimensional singly linked list: one for dervied type id and the other for IO id
# B) May use hash integer for derived type id
# C) IO id will be saved in each kw_ kr_ subroutine with SAVE attribute
# D) generate in (named?) data block
# E) Use module for dtype list and manipulate functions. Use Interface for assigning pointer

from collections import OrderedDict
from fortlab.resolver import statements, typedecl_statements, block_statements

shared_objects = OrderedDict()
shared_objects['driver_object'] = None

state_gencore_parts = OrderedDict()
state_gencore_contains = []

kernel_gencore_parts = OrderedDict()
kernel_gencore_contains = []

#### VARLIST
modreadsubrs = OrderedDict()
modwritesubrs = OrderedDict()
localread = []
localwrite = []

DRIVER_USE_PART = 'DUP'
DRIVER_DECL_PART = 'DDP'
DRIVER_EXEC_PART ='DEP'
DRIVER_CONTAINS_PART = 'DCP'
DRIVER_SUBP_PART = 'DSP'
DRIVER_ALLOC_PART = 'DAP'
DRIVER_DEALLOC_PART = 'DDAP'
DRIVER_READ_IN_ARGS = 'DRIA'
DRIVER_READ_IN_EXTERNS = 'DRIE'
DRIVER_CALLSITE_PART = 'DCP'
DRIVER_EVAL_ALLOC_PART = 'DEAP'
DRIVER_EVAL_DEALLOC_PART = 'DEDAP'
DRIVER_EVAL_READ_IN_ARGS = 'DERIA'
DRIVER_EVAL_READ_IN_EXTERNS = 'DERIE'
DRIVER_EVAL_CALLSITE_PART = 'DECP'
DRIVER_WARMUP_ALLOC_PART = 'DWAP'
DRIVER_WARMUP_DEALLOC_PART = 'DWDAP'
DRIVER_WARMUP_READ_IN_ARGS = 'DWRIA'
DRIVER_WARMUP_READ_IN_EXTERNS = 'DWRIE'
DRIVER_WARMUP_CALLSITE_PART = 'DWCP'

STATE_PBLOCK_USE_PART = 'SPBUP'
STATE_PBLOCK_DECL_PART = 'SPBDP'
STATE_PBLOCK_EXEC_PART ='SPBEP'
STATE_PBLOCK_CONTAINS_PART = 'SPBCP'
STATE_PBLOCK_SUBP_PART = 'SPBSP'
STATE_PBLOCK_WRITE_IN_ARGS = 'SPBWIA'
STATE_PBLOCK_WRITE_IN_EXTERNS = 'SPBWIE'
STATE_PBLOCK_WRITE_IN_LOCALS ='SPBWIL'
STATE_PBLOCK_WRITE_OUT_EXTERNS = 'SPBWOE'
STATE_PBLOCK_WRITE_OUT_LOCALS = 'SPBWOL'

STATE_TBLOCK_USE_PART = 'STBUP'
STATE_TBLOCK_DECL_PART = 'STBDP'
STATE_TBLOCK_CONTAINS_PART = 'STBCP'
STATE_TBLOCK_SUBP_PART = 'STBSP'

KERNEL_PBLOCK_USE_PART = 'KPBUP'
KERNEL_PBLOCK_DECL_PART = 'KPBDP'
KERNEL_PBLOCK_EXEC_PART ='KPBEP'
KERNEL_PBLOCK_CONTAINS_PART = 'KPBCP'
KERNEL_PBLOCK_SUBP_PART = 'KPBSP'
#KERNEL_PBLOCK_READ_IN_EXTERNS = 'KPBRIE'
KERNEL_PBLOCK_PREPROCESS = 'KPBPP'
KERNEL_PBLOCK_READ_IN_LOCALS ='KPBRIL'
KERNEL_PBLOCK_READ_OUT_EXTERNS = 'KPBROE'
KERNEL_PBLOCK_READ_OUT_LOCALS = 'KPBROL'
KERNEL_PBLOCK_BEFORE_KERNEL = 'KPBBK'
KERNEL_PBLOCK_AFTER_KERNEL = 'KPBAK'

KERNEL_TBLOCK_USE_PART = 'KTBUP'
KERNEL_TBLOCK_DECL_PART = 'KTBDP'
KERNEL_TBLOCK_CONTAINS_PART = 'KTBCP'
KERNEL_TBLOCK_SUBP_PART = 'KTBSP'

rprefix = 'kr'
wprefix = 'kw'
vprefix = 'kv'

MAXLEN_SUBPNAME =40

def varstr(name, attrstr, at=None): 

    if at is not None:
        return "%s : %s at module %s" % (name, attrstr, at)
    else:
        return "%s : %s" % (name, attrstr)

def get_ancestor_name(stmt, generation):
    assert stmt and hasattr(stmt, 'parent'), 'Given stmt does not have parent attribute.'
    assert isinstance(generation, int), 'Not integer type of generation.'

    ancestor = stmt.ancestors()[generation]
    if hasattr(ancestor, 'name'):
        return ancestor.name 
    else: return ''

def get_topname(stmt):
    return get_ancestor_name(stmt, 0)

def get_parentname(stmt):
    return get_ancestor_name(stmt, -1)

def get_dtype_subpname(typestmt):
    if not hasattr(get_dtype_subpname, 'kgen_subpname_cache'):
        get_dtype_subpname.kgen_subpname_cache = OrderedDict()

    assert typestmt, 'None type of typestmt'

    subpname = '%s_%s'%(get_topname(typestmt), typestmt.name)

    if len(subpname)<MAXLEN_SUBPNAME:
        return subpname
    else:
        if subpname in get_dtype_subpname.kgen_subpname_cache:
            return 'kgen_%s_typesubp%d'%(get_topname(typestmt), get_dtype_subpname.kgen_subpname_cache[subpname])
        else:
            subpindex = len(get_dtype_subpname.kgen_subpname_cache)
            get_dtype_subpname.kgen_subpname_cache[subpname] = subpindex
            return 'kgen_%s_typesubp%d'%(get_topname(typestmt), subpindex)

def get_typedecl_subpname(stmt, entity_name):
    if not hasattr(get_typedecl_subpname, 'kgen_subpname_cache'):
        get_typedecl_subpname.kgen_subpname_cache = OrderedDict()

    assert isinstance(stmt, typedecl_statements.TypeDeclarationStatement), 'None type of typedecl stmt'
    assert entity_name, 'No entity name is provided.'

    var = stmt.get_variable(entity_name)
    if var is None: return 'Unknown_name'

    prefix = [ get_parentname(stmt), stmt.name ] + list(stmt.selector)
    l = []
    if var.is_array(): l.append('dim%d'%var.rank)
    if var.is_pointer(): l.append('ptr')

    subpname = '_'.join(prefix+l)
    if len(subpname)<MAXLEN_SUBPNAME:
        return '_'.join(prefix+l)
    else:
        if subpname in get_typedecl_subpname.kgen_subpname_cache:
            return 'kgen_%s_subp%d'%(get_parentname(stmt), get_typedecl_subpname.kgen_subpname_cache[subpname])
        else:
            subpindex = len(get_typedecl_subpname.kgen_subpname_cache)
            get_typedecl_subpname.kgen_subpname_cache[subpname] = subpindex
            return 'kgen_%s_subp%d'%(get_parentname(stmt), subpindex)

def get_dtype_writename(typestmt):
    if typestmt is None: return
    subpname = get_dtype_subpname(typestmt)
    if subpname: return '%s_%s'%(wprefix, subpname)

def get_dtype_readname(typestmt):
    if typestmt is None: return
    subpname = get_dtype_subpname(typestmt)
    if subpname: return '%s_%s'%(rprefix, subpname)

def get_module_in_writename(modstmt):
    if modstmt is None: return
    return '%s_externs_in_%s'%(wprefix, modstmt.name)

def get_module_out_writename(modstmt):
    if modstmt is None: return
    return '%s_externs_out_%s'%(wprefix, modstmt.name)

def get_module_in_readname(modstmt):
    if modstmt is None: return
    return '%s_externs_in_%s'%(rprefix, modstmt.name)

def get_module_out_readname(modstmt):
    if modstmt is None: return
    return '%s_externs_out_%s'%(rprefix, modstmt.name)

def get_typedecl_writename(typestmt, entity_name):
    if typestmt is None: return
    subpname = get_typedecl_subpname(typestmt, entity_name)
    if subpname: return '%s_%s'%(wprefix, subpname)

def get_typedecl_readname(typestmt, entity_name):
    if typestmt is None: return
    subpname = get_typedecl_subpname(typestmt, entity_name)
    if subpname: return '%s_%s'%(rprefix, subpname)

def get_typedecl_verifyename(typestmt, entity_name):
    if typestmt is None: return
    subpname = get_typedecl_subpname(typestmt, entity_name)
    if subpname: return '%s_%s'%(vprefix, subpname)

def process_spec_stmts(stmt):
    if not stmt: return
    if not hasattr(stmt, 'spec_stmts'): return

    for spec_stmt in stmt.spec_stmts:
        node = spec_stmt.genkpair
        if not node: continue
        if not node.kgen_isvalid: continue
        if not hasattr(spec_stmt, 'geninfo') or len(spec_stmt.geninfo)==0: continue

        def is_uname(item, unames):
            import re
            iname = re.split(r'\(|\*|=', item)[0].strip()
            if iname in unames: return True
            else: return False

        if hasattr(spec_stmt, 'items'):
            new_items = []
            unames = list(set([ uname.firstpartname() for uname, req in KGGenType.get_state(spec_stmt.geninfo) ]))
            for item in spec_stmt.items:
                if is_uname(item, unames):
                    new_items.append(item)
            node.new_items = new_items
            node.kgen_use_tokgen = True
        else:
            pass
            # maybe specific handling per classes

def gen_write_istrue(pobj, var, ename):

    # if isarray

    if var.is_array():
        attrs = {'expr': 'SIZE(%s)==1'%ename}
        ifsizeobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'UBOUND(%s, 1)<LBOUND(%s, 1)'%(ename, ename)}
        ifarrobj = part_append_gensnode(ifsizeobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifarrobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'expr': 'UBOUND(%s, 1)==0 .AND. LBOUND(%s, 1)==0'%(ename, ename)}
        part_append_gensnode(ifarrobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifarrobj, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_gensnode(ifarrobj, EXEC_PART, block_statements.Else, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.TRUE.'}
        part_append_gensnode(ifarrobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'expr': 'SIZE(%s)==0'%ename}
        part_append_gensnode(ifsizeobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifsizeobj, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_gensnode(ifsizeobj, EXEC_PART, block_statements.Else, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.TRUE.'}
        part_append_gensnode(ifsizeobj, EXEC_PART, statements.Assignment, attrs=attrs)
    else:
        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.TRUE.'}
        part_append_gensnode(pobj, EXEC_PART, statements.Assignment, attrs=attrs)

    # if allocatable
    if var.is_allocatable():
        attrs = {'expr': '.NOT. ALLOCATED(%s)'%ename}
        ifallocobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifallocobj, EXEC_PART, statements.Assignment, attrs=attrs)

    # if pointer
    if var.is_pointer():
        attrs = {'expr': '.NOT. ASSOCIATED(%s)'%ename}
        ifptrobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifptrobj, EXEC_PART, statements.Assignment, attrs=attrs)


    if var.is_array() or var.is_allocatable() or var.is_pointer():

        attrs = {'items': ['kgen_istrue'], 'specs': ['UNIT = kgen_unit']}
        part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

        attrs = {'expr': 'kgen_istrue'}
        iftrueobj = part_append_gensnode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        pobj = iftrueobj

    if var.is_array() and var.parent.is_numeric():
        attrs = {'variable': 'kgen_array_sum', 'sign': '=', 'expr': 'DBLE(SUM(%s, mask=(%s .eq. %s)))'%(ename, ename, ename)}
        part_append_gensnode(pobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'items': ['kgen_array_sum'], 'specs': ['UNIT = kgen_unit']}
        part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)
    
    return pobj


def namedgen_write_istrue(kernel_id, partid, var, entity_name):

    pobj = None

    # if isarray
    if var.is_array():
        attrs = {'expr': 'SIZE(%s)==1'%entity_name}
        ifsizeobj = namedpart_append_gensnode(kernel_id, partid, block_statements.IfThen, attrs=attrs)

        attrs = {'expr': 'UBOUND(%s, 1)<LBOUND(%s, 1)'%(entity_name, entity_name)}
        ifarrobj = part_append_gensnode(ifsizeobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifarrobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'expr': 'UBOUND(%s, 1)==0 .AND. LBOUND(%s, 1)==0'%(entity_name, entity_name)}
        part_append_gensnode(ifarrobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifarrobj, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_gensnode(ifarrobj, EXEC_PART, block_statements.Else, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.TRUE.'}
        part_append_gensnode(ifarrobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'expr': 'SIZE(%s)==0'%entity_name}
        part_append_gensnode(ifsizeobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifsizeobj, EXEC_PART, statements.Assignment, attrs=attrs)

        part_append_gensnode(ifsizeobj, EXEC_PART, block_statements.Else, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.TRUE.'}
        part_append_gensnode(ifsizeobj, EXEC_PART, statements.Assignment, attrs=attrs)

    # if allocatable
    if var.is_allocatable():
        attrs = {'expr': '.NOT. ALLOCATED(%s)'%entity_name}
        ifallocobj = namedpart_append_gensnode(kernel_id, partid, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifallocobj, EXEC_PART, statements.Assignment, attrs=attrs)

    # if pointer
    if var.is_pointer():
        attrs = {'expr': '.NOT. ASSOCIATED(%s)'%entity_name}
        ifptrobj = namedpart_append_gensnode(kernel_id, partid, block_statements.IfThen, attrs=attrs)

        attrs = {'variable': 'kgen_istrue', 'sign': '=', 'expr': '.FALSE.'}
        part_append_gensnode(ifptrobj, EXEC_PART, statements.Assignment, attrs=attrs)


    if var.is_array() or var.is_allocatable() or var.is_pointer():

        attrs = {'items': ['kgen_istrue'], 'specs': ['UNIT = kgen_unit']}
        namedpart_append_gensnode(kernel_id, partid, statements.Write, attrs=attrs)

        attrs = {'expr': 'kgen_istrue'}
        iftrueobj = namedpart_append_gensnode(kernel_id, partid, block_statements.IfThen, attrs=attrs)

        pobj = iftrueobj

    if var.is_array() and var.parent.is_numeric():
        attrs = {'variable': 'kgen_array_sum', 'sign': '=', 'expr': 'DBLE(SUM(%s, mask=(%s .eq. %s)))'%(entity_name, entity_name, entity_name)}
        part_append_gensnode(pobj, EXEC_PART, statements.Assignment, attrs=attrs)

        attrs = {'items': ['kgen_array_sum'], 'specs': ['UNIT = kgen_unit']}
        part_append_gensnode(pobj, EXEC_PART, statements.Write, attrs=attrs)

    return pobj

def gen_read_istrue(subrobj, var, ename, allocate=False):

    pobj = subrobj
 
    if var.is_array() or var.is_allocatable() or var.is_pointer():
        attrs = {'items': ['kgen_istrue'], 'specs': ['UNIT = kgen_unit']}
        part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

        attrs = {'expr': 'kgen_istrue'}
        iftrueobj = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        pobj = iftrueobj

    if var.is_allocatable() or allocate:
        attrs = {'expr': 'ALLOCATED( %s )'%ename}
        ifalloc = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['%s'%ename]}
        part_append_genknode(ifalloc, EXEC_PART, statements.Deallocate, attrs=attrs)

    if var.is_pointer():
        attrs = {'expr': 'ASSOCIATED( %s )'%ename}
        ifalloc = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['%s'%ename]}
        part_append_genknode(ifalloc, EXEC_PART, statements.Nullify, attrs=attrs)

    if var.is_array() and var.parent.is_numeric():
        attrs = {'items': ['kgen_array_sum'], 'specs': ['UNIT = kgen_unit']}
        part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

    return pobj

def namedgen_read_istrue(kernel_id, partid, var, entity_name, ename_prefix=''):

    pobj = None

    if var.is_array() or var.is_allocatable() or var.is_pointer():
        attrs = {'items': ['kgen_istrue'], 'specs': ['UNIT = kgen_unit']}
        namedpart_append_genknode(kernel_id, partid, statements.Read, attrs=attrs)

        attrs = {'expr': 'kgen_istrue'}
        iftrueobj = namedpart_append_genknode(kernel_id, partid, block_statements.IfThen, attrs=attrs)

        pobj = iftrueobj

    if var.is_allocatable():
        attrs = {'expr': 'ALLOCATED( %s )'%(ename_prefix+entity_name)}
        ifalloc = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['%s'%(ename_prefix+entity_name)]}
        part_append_genknode(ifalloc, EXEC_PART, statements.Deallocate, attrs=attrs)

    if var.is_pointer():
        attrs = {'expr': 'ASSOCIATED( %s )'%(ename_prefix+entity_name)}
        ifalloc = part_append_genknode(pobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

        attrs = {'items': ['%s'%(ename_prefix+entity_name)]}
        part_append_genknode(ifalloc, EXEC_PART, statements.Nullify, attrs=attrs)

    if var.is_array() and var.parent.is_numeric():
        attrs = {'items': ['kgen_array_sum'], 'specs': ['UNIT = kgen_unit']}
        part_append_genknode(pobj, EXEC_PART, statements.Read, attrs=attrs)

    return pobj

def is_param_zero(length, stmt):
    if hasattr(stmt, 'unknowns'):
        for uname, req in stmt.unknowns.items():
            if uname.firstpartname()==length and len(req.res_stmts)>0:
                res_stmt = req.res_stmts[0]
                if isinstance(res_stmt, typedecl_statements.Integer) and 'parameter' in res_stmt.attrspec:
                    for decl in res_stmt.entity_decls:
                        vname, value = decl.split('=')            
                        if vname.strip()==length:
                            try:
                                intlen = int(value)
                                if intlen == 0: return True
                            except:
                                return is_param_zero(length, res_stmt)
    return False

def is_zero_array(var, stmt):
    # Temporary turn off
    #return False
    if var.is_explicit_shape_array():
        for length in var.shape:
            try:
                intlen = int(length)
                if intlen==0: return True
            except:
                if is_param_zero(length, stmt):
                    return True
    return False

def is_excluded(ename, stmt):
    if hasattr(stmt, 'exclude_names'):
        for name, actions in stmt.exclude_names.items():
            if ename==name: return True
    return False

def is_remove_state(ename, stmt):
    if hasattr(stmt, 'exclude_names'):
        for name, actions in stmt.exclude_names.items():
            if ename==name and 'remove_state' in actions:
                return True
    return False

def check_class_derived(stmt):

    if not stmt.is_class(): return False

    # check if the name of kind in unknows, and the last res of the unknown is derivied type
    dtypename = stmt.selector[1]
    if hasattr(stmt, 'unknowns'):
        for uname, req in stmt.unknowns.items():
            if uname.firstpartname()==dtypename and len(req.res_stmts)>0:
                if isinstance(req.res_stmts[0], (block_statements.Type, block_statements.TypeDecl)):
                    return True
    return False
